--[[
    Coconuts by LordFarlander
    Version 3.0.237
    $Revision: 237 $

    Used to be called Mount.
]]--

--[[
Copyright (c) 2008-2011, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Coconuts" );

_G["BINDING_NAME_CLICK Coconuts_Button:LeftButton"] = L["Summon Mount"];
BINDING_HEADER_COCONUTSHEADER = L["Coconuts"];

if( not LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-Skill-4.0" ) );
    return;
end--if

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true );
if( not LibLordFarlander ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-2.0" ) );
    return;
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true );
if( not ItemList ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-ItemList-2.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-UI-2.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-UI-2.0" ) );
    return;
end--if

local PetAndMountDatabase = LibStub( "LibLordFarlander-PetAndMountDatabase-1.1", true );
if( not PetAndMountDatabase ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-PetAndMountDatabase-1.1" ) );
    return;
end--if

local TR = LibStub( "LibLordFarlander-TableRecycler-1.0", true );
if( not TR ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-TableRecycler-1.0" ) );
    return;
end--if

local BZ = LibStub( "LibBabble-Zone-3.0", true ):GetLookupTable();

Coconuts = LibStub( "AceAddon-3.0" ):NewAddon( "Coconuts", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0" );
Coconuts.version = { major = 3, minor = 0, revision = tonumber( "237" ) };

Coconuts.DEBUG  = false;            --true = Debugging Mode
Coconuts.dprint = LibLordFarlander.dprint;
Coconuts.button = nil;
Coconuts.ui = { Favorites = TR:GetTable(), };

Coconuts.dirtyBags = { [-254] = true, [-255] = true, };
for bag = 0, NUM_BAG_SLOTS do
    Coconuts.dirtyBags[bag] = true;
end--for

local wowVersion;

do
    local v1, v2, v3 = strsplit( ".", GetBuildInfo() );
    
    wowVersion = tonumber( v1 ) * 10000 + tonumber( v2 ) * 100 + tonumber( v3 );
end--do

--DO NOT CHANGE THESE VALUES
Coconuts.Mounts = {
    Zones = TR:GetTable(),
    List = {
        Ground = {
            Normal = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
            Passenger = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
        },
        Flying = {
            Normal = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
            Passenger = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
        },
        Aquatic = {
            Normal = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
            Passenger = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
        },
        Underwater = {
            Normal = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
            Passenger = {
                Best = 0,
                NextBest = 0,
                Current = nil,
                NextCurrent = nil,
                Sets = { [0] = TR:GetTable( "speed", 0 ), },
            },
        },
    },
};

Coconuts.CurrentForm = { Ground = nil, Flying = nil, Aquatic = nil, };
Coconuts.CurrentForm.Flying  = nil;
Coconuts.CurrentForm.Aquatic = nil;
Coconuts.ReadyToInitialize  = false;
Coconuts.Mode               = "";
Coconuts.Loaded             = false;
Coconuts.PlayerState = { Mounted = false, Swimming = false, InVehicle = false, RidingSkill = nil, Has310 = false, WintergraspBattle = false };

Coconuts.MountItemNames = {
    [21212] = L["Fresh Holly"],
    [37750] = L["Fresh Brewfest Hops"],
    [39477] = L["Fresh Dwarven Hops"],
    [39476] = L["Fresh Goblin Hops"],
    [21213] = L["Preserved Holly"],
    [37816] = L["Preserved Brewfest Hops"],
};

local coldWeatherFlying = 54197;
local flightMastersLicence = 90267;

function Coconuts:OnInitialize()
    local AceConfigReg = LibStub( "AceConfigRegistry-3.0" );
    local AceConfigDialog = LibStub( "AceConfigDialog-3.0" );
    local LibLordFarlander_UI = LibStub( "LibLordFarlander-UI-2.0" );
    local defaults = {
            char = {
                FirstSeen = {},                
            },
            profile = {
                FavorFlying = true,
                FavorPassenger = true,
                ShowModifierItems = true,
                SuggestOnlySoulbound = true,
                ModiferDismountFlying = true,
                SuggestForms = true,
                FlightFormDefault = true,
                DefaultToFormInCombat = true,
                FavorMorePassengers = false,
                BoostNew = false,
                BoostNewTime = 86400 * 7, -- 1 week
                FlyingModifier = "Alt",
                NextBestModifier = "Shift",
                PassengerModifier = "Control",
                MountItemModifier = "Shift",
                -- List of favorite mounts, these will be picked more often
                -- stored as [itemID]=percentage
                Favorites = {},
                -- List of mounts per zone, if the mounts are in the player's bags or spells they will be picked
                -- when in the indicated zone
                -- stored as ["Localized zone name"]={ "type"=itemID, ... }
                ZoneFavorites = {},
                Button = LibLordFarlander_UI.GetButtonSettingsDefaults(),
            },
        };
    local buttonOptions = LibLordFarlander_UI.GetButtonOptionGroup( self, {
            name = L["Button"],
            desc = L["Options relating to the button."],
            yes = L["yes"],
            no = L["no"],
            show = L["showing"],
            hide = L["hiding"],
            locked = L["locked"],
            unlocked = L["unlocked"],
            toggleButton = L["toggleButton"],
            toggleButton_name = L["Show button"],
            toggleButton_desc = L["Show/hide the button."],
            autoHideButton = L["autoHideButton"],
            autoHideButton_name = L["Automatically hide button"],
            autoHideButton_desc = L["Hide the button when you can't mount?"],
            lockButton = L["lockButton"],
            lockButton_name = L["Lock button position"],
            lockButton_desc = L["Lock/unlock the position of the button."],
            alpha = L["alpha"],
            alpha_name = L["Alpha of button"],
            alpha_desc = L["Sets the alpha value that the button has normally."],
            fadeOut = L["fadeOut"],
            fadeOut_name = L["Fade out"],
            fadeOut_desc = L["Fade the button out when the cursor is not over it?"],
            fadeOutInsteadHide = L["fadeOutInsteadHide"],
            fadeOutInsteadHide_name = L["Fade out instead of hiding"],
            fadeOutInsteadHide_desc = L["Fade the button out instead of hiding it?"],
            fadeOutCancelInCombat = L["fadeOutCancelInCombat"],
            fadeOutCancelInCombat_name = L["Cancel fade out if in combat"],
            fadeOutCancelInCombat_desc = L["Should the button not fade out if you're in combat?"],
            fadeOutCancelOnShift = L["fadeOutCancelOnShift"],
            fadeOutCancelOnShift_name = L["Cancel fade out if holding Shift"],
            fadeOutCancelOnShift_desc = L["Should the button not fade out if you're holding Shift?"],
            fadeOutCancelOnCtrl = L["fadeOutCancelOnCtrl"],
            fadeOutCancelOnCtrl_name = L["Cancel fade out if holding Ctrl"],
            fadeOutCancelOnCtrl_desc = L["Should the button not fade out if you're holding Ctrl?"],
            fadeOutCancelOnAlt = L["fadeOutCancelOnAlt"],
            fadeOutCancelOnAlt_name = L["Cancel fade out if holding Alt"],
            fadeOutCancelOnAlt_desc = L["Should the button not fade out if you're holding Alt?"],
            fadeOutTime = L["fadeOutTime"],
            fadeOutTime_name = L["Fade out time"],
            fadeOutTime_desc = L["Sets how long the button takes to fade out."],
            fadeOutDelay = L["fadeOutDelay"],
            fadeOutDelay_name = L["Fade out delay"],
            fadeOutDelay_desc = L["Sets the delay before the button starts to fade."],
            fadeOutAlpha = L["fadeOutAlpha"];
            fadeOutAlpha_name = L["Alpha after fadeout"],
            fadeOutAlpha_desc = L["Sets the alpha value that the button has after faded out."],
            buttonScale = L["buttonScale"],
            buttonScale_name = L["Scale of button"],
            buttonScale_desc = L["Sets the scaling value for the button to make it bigger or smaller."],
        }, 1 );
    local modifierItemsOptions = {
        type = "group",
        name = L["Modifier Items"],
        desc = L["Allows you to set the order that modifier items are picked."],
        handler = self,
        args = {
            favorites = {
                type = "select",
                order = 1,
                name = L["Modifier Items"],
                desc = L["Allows you to set the order that modifier items are picked."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             local favorites = TR:GetTable();

                             for _, item in pairs( info.handler.db.profile.enchantOrder ) do
                                 local sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( item );

                                 if( sTexture ) then
                                     tinsert( favorites, ("|T%s:0|t %s"):format( sTexture, sName ) );
                                 else
                                     tinsert( favorites, info.handler.MountItemNames[item] or item );
                                 end--if
                             end--if
                             if( not info.handler.ui.currentModifierItem ) then
                                 info.handler.ui.currentModifierItem = 1;
                             end--if
                             return favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentModifierItem;
                      end,
                set = function( info, val )
                          info.handler.ui.currentModifierItem = val;
                      end,
            },
            up = {
                type = "execute",
                order = 2,
                name = L["Move up"],
                desc = L["Moves the selected modifier up in the list."],
                cmdHidden = true,
                func = function( info )
                           if( info.handler.ui.currentModifierItem > 1 ) then
                               local olditem = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem - 1];
                               
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem - 1] = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem];
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem] = olditem;
                               info.handler.ui.currentModifierItem = info.handler.ui.currentModifierItem - 1;
                               info.handler:FindNextMountEnchant();
                           end--if
                       end,
            },
            down = {
                type = "execute",
                order = 2,
                name = L["Move down"],
                desc = L["Moves the selected modifier down in the list."],
                cmdHidden = true,
                func = function( info )
                           if( info.handler.ui.currentModifierItem < #info.handler.db.profile.enchantOrder ) then
                               local olditem = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem + 1];
                               
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem + 1] = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem];
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem] = olditem;
                               info.handler.ui.currentModifierItem = info.handler.ui.currentModifierItem + 1;
                               info.handler:FindNextMountEnchant();
                           end--if
                       end,
            },
        },
    };
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Mounts that should be picked more often if you have them."],
        handler = self,
        args = {
            favorites = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Mounts that should be picked more often if you have them."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             info.handler.ui.Favorites = table.wipe( info.handler.ui.Favorites );

                             local Favorites = info.handler.ui.Favorites;
                             local firstItem = nil;
                             local used = TR:GetTable();

                             for _, mount in pairs( info.handler:GetAllMountsList() ) do
                                 local id = mount:GetID();

                                 used[id] = true;
                                 table.insert( Favorites, id );
                                 if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                     info.handler.db.profile.Favorites[id] = 2.0;
                                 end--if
                                 if( ( type( info.handler.db.profile.Favorites[id] ) ~= "nil" ) and ( info.handler.db.profile.Favorites[id] == 1.0 ) ) then
                                     info.handler.db.profile.Favorites[id] = nil;
                                 end--if
                             end--for
                             for id, howOften in pairs( info.handler.db.profile.Favorites ) do
                                 if( not used[id] and ( ( id < -1 ) and GetSpellInfo( id * -1 ) or GetItemInfo( id ) ) ) then
                                     table.insert( Favorites, id );
                                     used[id] = true;
                                 end--if
                             end--if
                             TR:ReturnTable( used );
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = TR:GetTable();
                             for index, mount in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;
                                 local sHowOften = L["default"];
                                 local howOften = info.handler.db.profile.Favorites[mount] or 1.0;

                                 if( mount < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                                 end--if
                                 if( howOften == 0.0 ) then
                                     sHowOften = L["never"];
                                 elseif( howOften ~= 1.0 ) then
                                     sHowOften = L["%.1f times as often"]:format( howOften );
                                 end--if                                 
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                                 end--if
                                 if( firstItem == nil ) then
                                     firstItem = index;
                                 end--if
                             end--for
                             if( ( not Favorites[info.handler.ui.currentFavorite] ) or ( info.handler.ui.currentFavorite == nil ) ) then
                                 info.handler.ui.currentFavorite = firstItem;
                             end--if
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                      end,
            },
            frequency = {
                type = "range",
                order = 2,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the mount gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          return info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] or 1.0;
                      end,
                set = function( info, val )
                          if( val == 1.0 ) then
                              info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          else
                              info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = val;
                          end--if
                      end,
            },

            [L["list"]] = {
                type = "execute",
                order = 3,
                guiHidden = true,
                name = L["List all favorite mounts"],
                desc = L["Show a list with all of your favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your favorite mounts:"] );
                           for mount, howOften in pairs( info.handler.db.profile.Favorites ) do
                               if( howOften ~= 1.0 ) then
                                   local link = ( mount > 0 ) and LibLordFarlander.GetItemLink( mount ) or GetSpellLink( mount * -1 );

                                   if( mount > 0 ) then
                                       sTexture = select( 10, GetItemInfo( mount ) );
                                   else
                                       sTexture = select( 3, GetSpellInfo( mount * -1 ) );
                                   end--if
                                   if( howOften == 0.0 ) then
                                       howOften = L["never"];
                                   else
                                       howOften = L["%.1f times as often"]:format( howOften );
                                   end--if

                                   if( sTexture ) then
                                       info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( mount ), howOften ) );
                                   else
                                       info.handler:Print( ("%s"):format( link or tostring( mount ) ) );
                                   end--if
                               end--if
                           end--for
                       end,
            },
        },
    };
    local typeTemplate = {
        header = {
            type = "header",
            order = 1,
            width = "full",
            cmdHidden = true,
            name = function( info )
                       local setname = LibLordFarlander.UnnamifyString( info[#info - 1] );
                       local speed = select( 3, strsplit( ".", setname ) );
                       local mounttype;

                       if( setname:find( "Passenger" ) ) then
                           if( setname:find( "Flying" ) ) then
                               mounttype = L["flying passenger"];
                           elseif( setname:find( "Aquatic" ) ) then
                               mounttype = L["swimming passenger"];
                           elseif( setname:find( "Underwater" ) ) then
                                mounttype = L["underwater passenger"];
                           else
                               mounttype = L["ground passenger"];
                           end--if
                       else
                           if( setname:find( "Flying" ) ) then
                               mounttype = L["flying"];
                           elseif( setname:find( "Aquatic" ) ) then
                               mounttype = L["swimming"];
                           elseif( setname:find( "Underwater" ) ) then
                               mounttype = L["underwater"];
                           else
                               mounttype = L["ground"];
                           end--if
                       end--if
                       return ("%s: %s"):format( LibLordFarlander.UnnamifyString( info[#info - 2] ), L["%1$i%% speed %2$s"]:format( speed - 100, mounttype ) );
                   end,
        },
        mount = {
            type = "select",
            order = 2,
            name = L["Mount"],
            desc = L["Mounts that should be picked more often if you have them in the selected zone and type."],
            width = "full",
            cmdHidden = true,
            values = function( info )
                         local currentType = LibLordFarlander.UnnamifyString( info[#info - 1] ) or "";
                         local currentZone = LibLordFarlander.UnnamifyString( info[#info - 2] );

                         info.handler.ui.Favorites = table.wipe( info.handler.ui.Favorites );
                         if( not info.handler.db.profile.ZoneFavorites[currentZone] ) then
                             info.handler.db.profile.ZoneFavorites[currentZone] = TR:GetTable();
                         end--if

                         local Favorites = info.handler.ui.Favorites;
                         local used = TR:GetTable();
                         local ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];
                         local mountType, subType, speed = strsplit( ".", currentType );
                         
                         speed = tonumber( speed );
                         for _, mount in pairs( info.handler:GetPetsList( speed, info.handler.Mounts.List[mountType][subType] ) ) do
                             local id = mount:GetID();
                             table.insert( Favorites, id );
                             used[id] = true;
                             if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                 info.handler.db.profile.Favorites[id] = 2.0;
                             end--if
                             if( ( type( info.handler.db.profile.Favorites[id] ) ~= "nil" ) and ( info.handler.db.profile.Favorites[id] == 1.0 ) ) then
                                 info.handler.db.profile.Favorites[id] = nil;
                             end--if                             
                         end--for
                         if( ZoneFavorite ) then
                             if( type( ZoneFavorite ) ~= "table" ) then
                                 info.handler.db.profile.ZoneFavorites[currentZone][currentType] = { [ZoneFavorite] = { frequency = 1.0, exclusive = true } };
                                 ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];
                             end--if
                             for mount, _ in pairs( ZoneFavorite ) do
                                 if( not used[mount] and ( ( mount < -1 ) and GetSpellInfo( mount * -1 ) or GetItemInfo( mount ) ) ) then
                                     table.insert( Favorites, mount );
                                     used[mount] = true;
                                 end--if
                             end--for
                         else
                             info.handler.db.profile.ZoneFavorites[currentZone][currentType] = TR:GetTable();
                             ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];
                         end--if
                         TR:ReturnTable( used );
                         table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                         Favorites = TR:GetTable();
                         for index, mount in pairs( info.handler.ui.Favorites ) do
                             local sName, sTexture;
                             local howOften = 1.0;
                             local sHowOften = L["default"]
                             local exclusive = false;

                             if( ZoneFavorite[mount] ) then
                                 howOften = ZoneFavorite[mount].frequency or 1.0;
                                 exclusive = ZoneFavorite[mount].exclusive;
                             end--if
                             if( mount < 0 ) then
                                 sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                             else
                                 sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                             end--if
                             if( howOften == 0.0 ) then
                                 sHowOften = L["never"];
                             elseif( howOften ~= 1.0 ) then
                                 sHowOften = L["%.1f times as often"]:format( howOften );
                             end--if
                             if( exclusive ) then
                                 sHowOften = string.format( "%s, %s", sHowOften, L["exclusive"] );
                             end--if
                             if( sTexture ) then
                                 Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                             end--if
                         end--for
                         if( #Favorites == 0 ) then
                             info.handler.ui.currentFavorite = nil;
                             Favorites[0] = L["none"];
                         elseif( ( info.handler.ui.currentFavorite == nil ) or ( info.handler.ui.currentFavorite > #Favorites ) ) then
                             info.handler.ui.currentFavorite = 1;
                         end--if
                         return Favorites;
                     end,
            get = function( info )
                      return info.handler.ui.currentFavorite;
                  end,
            set = function( info, val )
                      info.handler.ui.currentFavorite = val;
                  end,
            },
            frequency = {
                type = "range",
                order = 3,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the mount gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];
                          
                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];

                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].frequency or 1.0;
                              end--if
                          end--if
                          return 1.0;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( val == 1.0 ) then
                              val = nil;
                          end--if
                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = TR:GetTable();
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = TR:GetTable();
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = TR:GetTable();
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { frequency = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.frequency = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            active = {
                type = "toggle",
                order = 4,
                name = L["Override favorites"],
                desc = L["Check this to override the favorites settings for this mount."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];
                          
                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];
                              
                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].active or false;
                              end--if
                          end--if
                          return false;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = TR:GetTable();
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = TR:GetTable();
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = TR:GetTable();
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { active = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.active = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            exclusive = {
                type = "toggle",
                order = 5,
                name = L["Exclusive"],
                desc = L["Mounts marked Exclusive will be the only ones picked from."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];

                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];

                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].exclusive or false;
                              end--if
                          end--if
                          return false;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = TR:GetTable();
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = TR:GetTable();
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = TR:GetTable();
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { exclusive = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.exclusive = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
    };
    local zoneTemplate = {
    };
    local continentTemplate = {
    };
    local zoneOptions =  {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a mount to use when in a certain zone."],
        handler = self,
        childGroups = "tree",
        args = {
            [L["list"]] = {
                type = "execute",
                order = -1,
                guiHidden = true,
                name = L["List all zone favorite mounts"],
                desc = L["Show a list with all of your zone favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your zone favorite mounts:"] );
                           for zone, types in pairs( info.handler.db.profile.ZoneFavorites ) do
                               for type, mounts in pairs( types ) do
                                   for mount, settings in pairs( mounts ) do
                                       local howOften = settings.frequency or 1.0;
                                       local sHowOften =  "default";
                                       local exclusive = settings.exclusive;

                                       if( ( howOften ~= 1.0 ) or exclusive ) then
                                           local link = ( mount > 0 ) and LibLordFarlander.GetItemLink( mount ) or GetSpellLink( mount * -1 );

                                           if( mount > 0 ) then
                                               sTexture = select( 10, GetItemInfo( mount ) );
                                           else
                                               sTexture = select( 3, GetSpellInfo( mount * -1 ) );
                                           end--if
                                           if( howOften == 0.0 ) then
                                               sHowOften = L["never"];
                                           elseif( howOften ~= 1.0 ) then
                                               sHowOften = L["%.1f times as often"]:format( howOften );
                                           end--if
                                           if( exclusive ) then
                                               sHowOften = ("%s, %s"):format( sHowOften, L["exclusive"] );
                                           end--if
                                           if( sTexture ) then
                                               info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( mount ), sHowOften ) );
                                           else
                                               info.handler:Print( ("%s |cFFFF0000[%s]|r"):format( link or tostring( mount ), sHowOften ) );
                                           end--if
                                       end--if
                                   end--for
                               end--for
                           end--for
                       end,
            },
        },
    };
    local modifierKeys = { ["Disabled"] = L["Disabled"], ["Alt"] = L["Alt"], ["Control"] = L["Control"], ["Shift"] = L["Shift"], };
    local options = {
        type = "group",
        handler = self,
        args = {
            [L["favorFlying"]] = {
                type = "toggle",
                order = 1,
                name = L["Favor flying mounts"],   
                arg = {
                    DescriptionString = L["Toggles if flying mounts should be favored if in an area that allows them."],
                    VarableName = "FavorFlying",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                         
            },
            [L["favorPassenger"]] = {
                type = "toggle",
                order = 2,
                name = L["Favor passenger mounts"],
                arg = {
                    DescriptionString = L["Toggles if mounts that can take a passenger should be favored."],
                    VarableName = "FavorPassenger",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                        
            },
            [L["favorMorePassengers"]] = {
                type = "toggle",
                order = 3,
                name = L["Favor more passengers"],
                arg = {
                    DescriptionString = L["Toggles if mounts that can take more passengers should be favored over mounts that can take less."],
                    VarableName = "FavorMorePassengers",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                       
            },
            [L["showModifierItems"]] = {
                type = "toggle",
                order = 4,
                name = L["Show modifier items"],
                arg = {
                    DescriptionString = L["Toggles if items that modify mounts (such as Fresh Holly) should be selected when holding Ctrl while mounted."],
                    VarableName = "ShowModifierItems",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = LibLordFarlander_UI.SetGenericOption,                       
            },
            [L["modiferDismountFlying"]] = {
                type = "toggle",
                order = 6,
                name = L["Need modifer to dismount while flying"],
                width = "double",
                arg = {
                    DescriptionString = L["Toggles if a modifier needs to be held down to dismount while flying."],
                    VarableName = "ModiferDismountFlying",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                         
            },
            [L["suggestForms"]] = {
                type = "toggle",
                order = 7,
                name = L["Suggest travel forms instead of next best mount"],
                width = "double",
                arg = {
                    DescriptionString = L["Toggles if a travel form spell will be selected instead of your next slowest mount, if availiable."],
                    VarableName = "SuggestForms",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                      
            },
            [L["flightFormDefault"]] = {
                type = "toggle",
                order = 8,
                name = L["Default to flight form"],
                width = "double",
                arg = {
                    DescriptionString = L["Toggles if flight form will be defaulted to instead of a flying mount."],
                    VarableName = "FlightFormDefault",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                        
            },
            [L["defaultToFormInCombat"]] = {
                type = "toggle",
                order = 9,
                name = L["Suggest travel form when in combat"],
                width = "double",                      
                arg = {
                    DescriptionString = L["Toggles if when entering combat your travel form spell will be selected."],
                    VarableName = "DefaultToFormInCombat",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = LibLordFarlander_UI.SetGenericOption,                      
            },
            [L["boostNew"]] = {
                type = "toggle",
                order = 10,
                name = L["Boost new mounts"],
                width = "double",                    
                arg = {
                    DescriptionString = L["Toggles if new mounts will be selected more often than old ones."],
                    VarableName = "BoostNew",
                },
                desc = LibLordFarlander_UI.BooleanOptionDescription,
                get = LibLordFarlander_UI.GetGenericOption,
                set = "SetOption_ResetMounts",                      
            },
            BoostNewTime = {
                type = "range",
                order = 11,
                name = L["New time"],
                cmdHidden = true,
                min = 0.0,
                max = 31.0,
                bigStep = 1.0,
                desc = L["Set for how many days mounts are concidered new."];
                get = function( info )
                          return info.handler.db.profile.BoostNewTime / 86400;
                      end,
                set = function( info, val )
                          info.handler.db.profile.BoostNewTime = val * 86400;
                      end,
            },
            flyingModifier = {
                type = "select",
                order = 10,
                name = L["Modifier key for flying mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non flying mount in an area where flying is allowed."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.FlyingModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FlyingModifier = val;
                      end,
            },
            passengerModifier = {
                type = "select",
                order = 11,
                name = L["Modifier key for passenger mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non passenger mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.PassengerModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.PassengerModifier = val;
                      end,
            },
            mountItemModifier = {
                type = "select",
                order = 12,
                name = L["Modifier key for modifier items"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a modifer item while mounted."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.MountItemModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.MountItemModifier = val;
                      end,
            },
            nextBestModifier = {
                type = "select",
                order = 13,
                name = L["Modifier key for next slowest mount"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick your next slowest mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.NextBestModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.NextBestModifier = val;
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 14,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                cmdHidden = true,
                func = "ListMounts",
            },
            [L["refresh"]] = {
                type = "execute",
                order = 15,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                cmdHidden = true,
                func = "ResetMounts",
            },
        },
        name = L["Coconuts"],
        desc = L["Addon options."],
        icon = "Interface/AddOns/Coconuts/Img/Coconuts",
    };

    self.db = LibStub( "AceDB-3.0" ):New( "CoconutsDB", defaults );
    self.db.RegisterCallback( self, "OnProfileChanged" );
    self.db.RegisterCallback( self, "OnProfileReset" );

    local profileOptions = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( self.db );
    local slashOptions = {
        type = "group",
        handler = self,
        name = options.name,
        desc = options.desc,
        icon = options.icon,
        args = {
            [L["options"]] = options,
            [L["favorites"]] = favoritesOptions,
            [L["zoneFavorites"]] = zoneOptions,
            [L["button"]] = buttonOptions,
            [L["profile"]] = profileOptions,
            [L["gui"]] = {
                type = "execute",
                order = -1,
                name = L["Open GUI configurator"],
                desc = L["Open GUI configurator"],
                guiHidden = true,
                func = function( info )
                           InterfaceOptionsFrame_OpenToCategory( L[info.handler.name] );
                       end,
            },
            [L["newRandom"]] = {
                type = "execute",
                order = 6,
                name = L["Force the picking of new random mounts"],
                desc = L["Force the picking of new random mounts"],
                guiHidden = true,
                func = function( info )
                           info.handler:ChooseRandomMounts();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
            [L["nextItem"]] = {
                type = "execute",
                order = 7,
                name = L["Cycle to the next mount modifier item"],
                desc = L["Cycle to the next mount modifier item"],
                guiHidden = true,
                func = function( info )
                           info.handler:FindNextMountEnchant();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
            [L["summon"]] = {
                type = "group",
                order = 8,
                name = L["Summoning mounts"],
                desc = L["Summoning mounts"],
                guiHidden = true,
                args = {
                    [L["best"]] = {
                        type = "execute",
                        order = 1,
                        name = L["Summons the current best random mount (if a learned companion)"],
                        desc = L["Summons the current best random mount (if a learned companion)"],
                        guiHidden = true,
                        func = function( info )
                                   local mount = nil;
                                   local CurrentMountIndex = nil;

                                   -- logic to find the best mount
                                   if( LibLordFarlander.IsInFlyableZone() ) then
                                       if( profile.FavorFlying ) then
                                           if( profile.FavorPassenger ) then
                                               if( self.CurrentFlyingPassenger ) then
                                                   mount = self.CurrentFlyingPassenger;
                                               end--if
                                           end--if
                                           if( self.CurrentFlying ) then
                                               mount = self.CurrentFlying;
                                           end--if
                                       end--if
                                   end--if
                                   if( profile.FavorPassenger ) then
                                       if( self.CurrentGroundPassenger ) then
                                           mount = self.CurrentGroundPassenger;
                                       end--if
                                   end--if
                                   if( self.CurrentGround ) then
                                       mount = self.CurrentGround;
                                   end--if
                                   if( mount ) then
                                       info.handler:QuickSummon( mount.spellID );
                                   end--if
                               end,
                    },
                    [L["swimming"]] = {
                        type = "group",
                        order = 2,
                        name = L["Swimming mounts"],
                        desc = L["Swimming mounts"],
                        guiHidden = true,
                        args = {
                            [L["best"]] = {
                                type = "execute",
                                order = 1,
                                name = L["Summons the current best random mount (if a learned companion)"],
                                desc = L["Summons the current best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentGround ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Aquatic.Normal.Current.spellID );
                                           end--if
                                       end,
                            },
                            [L["nextbest"]] = {
                                type = "execute",
                                order = 2,
                                name = L["Summons the current next best random mount (if a learned companion)"],
                                desc = L["Summons the current next best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.NextCurrentGround ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Aquatic.Normal.NextCurrent.spellID );
                                           end--if
                                       end,
                            },
                        },
                    },
                    [L["ground"]] = {
                        type = "group",
                        order = 2,
                        name = L["Ground mounts"],
                        desc = L["Ground mounts"],
                        guiHidden = true,
                        args = {
                            [L["best"]] = {
                                type = "execute",
                                order = 1,
                                name = L["Summons the current best random mount (if a learned companion)"],
                                desc = L["Summons the current best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentGround ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Ground.Normal.Current.spellID );
                                           end--if
                                       end,
                            },
                            [L["nextbest"]] = {
                                type = "execute",
                                order = 2,
                                name = L["Summons the current next best random mount (if a learned companion)"],
                                desc = L["Summons the current next best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.NextCurrentGround ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Ground.Normal.NextCurrent.spellID );
                                           end--if
                                       end,
                            },
                        },
                    },
                    [L["flying"]] = {
                        type = "group",
                        order = 3,
                        name = L["Flying mounts"],
                        desc = L["Flying mounts"],
                        guiHidden = true,
                        args = {
                            [L["best"]] = {
                                type = "execute",
                                order = 1,
                                name = L["Summons the current best random mount (if a learned companion)"],
                                desc = L["Summons the current best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentFlying ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Flying.Normal.Current.spellID );
                                           end--if
                                       end,
                            },
                            [L["nextbest"]] = {
                                type = "execute",
                                order = 2,
                                name = L["Summons the current next best random mount (if a learned companion)"],
                                desc = L["Summons the current next best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentFlying ) then
                                               info.handler:QuickSummon( info.handler.List.Mounts.Flying.Normal.NextCurrent.spellID );
                                           end--if
                                       end,
                            },                        
                        },
                    },
                    [L["passenger"]] = {
                        type = "group",
                        order = 4,
                        name = L["Passenger mounts"],
                        desc = L["Passenger mounts"],
                        guiHidden = true,
                        args = {
                            [L["ground"]] = {
                                type = "group",
                                order = 2,
                                name = L["Ground mounts"],
                                desc = L["Ground mounts"],
                                guiHidden = true,
                                args = {
                                    [L["best"]] = {
                                        type = "execute",
                                        order = 1,
                                        name = L["Summons the current best random mount (if a learned companion)"],
                                        desc = L["Summons the current best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Ground.Passenger.Current ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Ground.Passenger.Current.spellID );
                                                   end--if
                                               end,
                                    },
                                    [L["nextbest"]] = {
                                        type = "execute",
                                        order = 2,
                                        name = L["Summons the current next best random mount (if a learned companion)"],
                                        desc = L["Summons the current next best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Ground.Passenger.NextCurrent ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Ground.Passenger.NextCurrent.spellID );
                                                   end--if
                                               end,
                                    },
                                },
                            },
                            [L["flying"]] = {
                                type = "group",
                                order = 2,
                                name = L["Flying mounts"],
                                desc = L["Flying mounts"],
                                guiHidden = true,
                                args = {
                                    [L["best"]] = {
                                        type = "execute",
                                        order = 1,
                                        name = L["Summons the current best random mount (if a learned companion)"],
                                        desc = L["Summons the current best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Flying.Passenger.Current ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Flying.Passenger.Current.spellID );
                                                   end--if
                                               end,
                                    },
                                    [L["nextbest"]] = {
                                        type = "execute",
                                        order = 2,
                                        name = L["Summons the current next best random mount (if a learned companion)"],
                                        desc = L["Summons the current next best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Flying.Passenger.NextCurrent ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Flying.Passenger.NextCurrent.spellID );
                                                   end--if
                                               end,
                                    },
                                },
                            },
                            [L["swimming"]] = {
                                type = "group",
                                order = 2,
                                name = L["Swimming mounts"],
                                desc = L["Swimming mounts"],
                                guiHidden = true,
                                args = {
                                    [L["best"]] = {
                                        type = "execute",
                                        order = 1,
                                        name = L["Summons the current best random mount (if a learned companion)"],
                                        desc = L["Summons the current best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Aquatic.Passenger.Current ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Aquatic.Passenger.Current.spellID );
                                                   end--if
                                               end,
                                    },
                                    [L["nextbest"]] = {
                                        type = "execute",
                                        order = 2,
                                        name = L["Summons the current next best random mount (if a learned companion)"],
                                        desc = L["Summons the current next best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.List.Mounts.Aquatic.Passenger.NextCurrent ) then
                                                       info.handler:QuickSummon( info.handler.List.Mounts.Aquatic.Passenger.NextCurrent.spellID );
                                                   end--if
                                               end,
                                    },
                                },
                            },
                        },
                    },
                },
            },
            [L["list"]] = {
                type = "execute",
                order = 14,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                guiHidden = true,
                func = "ListMounts",
            },
            [L["refresh"]] = {
                type = "execute",
                order = 15,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                guiHidden = true,
                func = "ResetMounts",
            },
            debug = {
                type = "toggle",
                hidden = true,
                name = "Debug",
                desc = "Debug mode toggle",
                desc = function( info )
                            local returnString = "Debug mode toggle";

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.DEBUG and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.DEBUG;
                      end,
                set = function( info, val )
                          info.handler.DEBUG = val;
                          info.handler:Print( ("debug mode |cFFFF0000[%s]|r"):format( info.handler.DEBUG and L["yes"] or L["no"] ) );
                      end,
            },
        },
    };
    local function ProcessSets( addTo, sets, mainName, subName, groupName, indexAdd )
        if( sets ) then
            for _, name in pairs( sets ) do
                local speed = self.GetSetSpeed( name );
                
                if( speed ) then
                    local index = speed + indexAdd;
                    local sType = L["%1$i%% speed %2$s"]:format( speed, groupName );

                    addTo[LibLordFarlander.NamifyString( ("%s.%s.%i"):format( mainName, subName, speed ) )] = {
                        type = "group",
                        handler = self,
                        order = index,
                        name = sType,
                        desc = sType,
                        args = typeTemplate,
                    };
                end--if
            end--for
        end--if
    end--ProcessSets( sets )
    local groundSets = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
    local groundPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" );
    local flyingSets = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );
    local flyingPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" );
    local aquaticSets = PetAndMountDatabase:GetSubsets( "Mounts.Aquatic" );
    local aquaticPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Aquatic" );
    local underwaterSets = PetAndMountDatabase:GetSubsets( "Mounts.Underwater" );
    local underwaterPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Underwater" );

    for continentindex, continent in pairs( { GetMapContinents() } ) do
        zoneOptions.args[LibLordFarlander.NamifyString( continent )] = {
            type = "group",
            handler = self,
            name = continent,
            desc = L["Zones in %s"]:format( continent ),
            childGroups = "tree",
            args = TR:GetTable(),
        };

        local continentArgs = zoneOptions.args[LibLordFarlander.NamifyString( continent )].args;

        LibLordFarlander.CopyTable( continentTemplate, continentArgs );
        for index, zone in pairs( LibLordFarlander.GetMapZonesWithExtra( continentindex, true ) ) do
            continentArgs[LibLordFarlander.NamifyString( zone )] = {
                type = "group",
                handler = self,
                name = zone,
                desc = zone,
                childGroups = "tree",
                args = TR:GetTable(),
            };
            
            local zoneArgs = continentArgs[LibLordFarlander.NamifyString( zone )].args;

            LibLordFarlander.CopyTable( zoneTemplate, zoneArgs );
            ProcessSets( zoneArgs, groundSets, "Ground", "Normal", L["ground"], 0 );
            ProcessSets( zoneArgs, groundPassengerSets, "Ground", "Passenger", L["ground passenger"], 1000 );
            ProcessSets( zoneArgs, aquaticSets, "Aquatic", "Normal", L["swimming"], 20000 );
            ProcessSets( zoneArgs, aquaticPassengerSets, "Aquatic", "Passenger", L["swimming passenger"], 21000 );
            ProcessSets( zoneArgs, underwaterSets, "Underwater", "Normal", L["underwater"], 30000 );
            ProcessSets( zoneArgs, underwaterPassengerSets, "Underwater", "Passenger", L["underwater passenger"], 31000 );
            if( LibLordFarlander.IsFlyableZone( zone, true ) ) then
                ProcessSets( zoneArgs, flyingSets, "Flying", "Normal", L["flying"], 10000 );
                ProcessSets( zoneArgs, flyingPassengerSets, "Flying", "Passenger", L["flying passenger"], 11000 );
            end--if
        end--for
    end--for

    LibStub( "AceConfig-3.0" ):RegisterOptionsTable( L["Coconuts"], slashOptions, { L["coconuts"], L["ccn"] } );
    AceConfigReg:RegisterOptionsTable( "Coconuts Main Options", options );
    AceConfigReg:RegisterOptionsTable( "Coconuts Button", buttonOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Zone Favorites", zoneOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Favorites", favoritesOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Profile", profileOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Modifer Items", modifierItemsOptions );
    AceConfigDialog:AddToBlizOptions( "Coconuts Main Options", L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Button", L["Button"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Modifer Items", L["Modifier Items"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Zone Favorites", L["Zone Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Favorites", L["Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Profile", L["Profile"], L["Coconuts"] );
    
    if( LibStub( "LibDataBroker-1.1", true ) ) then
        self.ldbobj = LibStub( "LibDataBroker-1.1" ):NewDataObject( L["Coconuts"], {
                type = "data source",
                icon = [[Interface\AddOns\Coconuts\Img\Coconuts]],
                label = L["Coconuts"],
                text = L["You have no mounts."],
                secureTemplate = "SecureActionButtonTemplate",
                OnClick = function( button, btn )
                              if( button:IsObjectType( "SecureActionButtonTemplate" ) ) then
                                  self.button:Click( btn );
                              else
                                  if( self.mode == "Enchant" ) then
                                      self:FindNextMountEnchant();
                                  elseif( self.mode == "Dismount" ) then
                                      if( ( not IsFlying() ) or ( self.db.char.ModiferDismountFlying and (
                                              IsAltKeyDown() or IsControlKeyDown() or IsShiftKeyDown() ) or ( not self.db.char.ModiferDismountFlying ) ) ) then
                                          Dismount();
                                      end--if
                                  elseif( self.mode == "VehicleExit" ) then
                                      VehicleExit();
                                  else
                                      if( self.CurrentMountIndex ) then
                                          CallCompanion( "MOUNT", self.CurrentMountIndex );
                                      end--if
                                      self:ChooseRandomMounts();
                                  end--if
                                  self:SetButton( self:GetStateBasedOnModifiers() );
                              end--if
                          end,
                OnTooltipShow = function( tooltip )
                                    if( not tooltip ) then
                                        return;
                                    end--if
                                    self:SetTooltip( tooltip );
                                end,
            } );
    end--if
end--Coconuts:OnInitialize()

function Coconuts:OnProfileReset()
    self:OnProfileChanged();
end--Coconuts:OnProfileReset()

function Coconuts:OnProfileChanged()
    LibStub( "LibLordFarlander-UI-2.0" ).OnProfileChanged( self );
    profile = self.db.profile;
    if( profile.enchantOrder ) then
        local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
        local item, value = iterator( handle );

        while item do
            local inList = false;

            for _, listItem in pairs( profile.enchantOrder ) do
                if( listItem == item ) then
                    inList = true;
                    break;
                end--if
            end--for
            if( not inList ) then
                tinsert( profile.enchantOrder, item );
            end--if
            item, value = iterator( handle );
        end--for
    else
        self.db.profile.enchantOrder = TR:GetTable();    

        local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
        local item, value = iterator( handle );
        local found = false;
        local counter = 1;

        while item do
            if( type( value ) == "boolean" ) then
                value = counter;
            else
                value = tonumber( value );
            end--if
            profile.enchantOrder[value] = item;
            item, value = iterator( handle );
            counter = counter + 1;
        end--for
    end--if
    --Upgrade check
    if( self.db.profile.FirstSeen ) then
        self.db.char.FirstSeen = self.db.profile.FirstSeen;
        self.db.profile.FirstSeen = nil;
    end--if    
    if( type( profile.LastVersion ) == "number" ) then
        --Upgrading from pre-2.5, need to convert zone favorites to the new format
        local newList = TR:GetTable();
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local GroundSets = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
        local FlyingSets = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );

        for zone, itemID in pairs( profile.ZoneFavorites ) do
            newList[zone] = TR:GetTable();
            for _, set in pairs( GroundSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    local speed = 0;

                    if( set:find( "Slow" ) ) then
                        speed = 0;
                    elseif( set:find( "Fast" ) ) then
                        speed = 60;
                    elseif( set:find( "VeryFast" ) ) then
                        speed = 100;
                    end--if
                    newList[zone][("Ground.Normal.%i"):format( speed )] = { itemID = { exclusive = true, } };
                end--if
            end--if
            for _, set in pairs( FlyingSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    local speed = 0;

                    if( set:find( "Slow" ) ) then
                        speed = 0;
                    elseif( set:find( "Fast" ) ) then
                        speed = 60;
                    elseif( set:find( "VeryFast" ) ) then
                        speed = 200;
                    elseif( set:find( "ExtremelyFast" ) ) then
                        speed = 310;
                    end--if
                    newList[zone][("Flying.Normal.%i"):format( speed )] = { itemID = { exclusive = true, } };
                end--if
            end--if
        end--for
        profile.ZoneFavorites = newList;
    end--if
    if( ( type( profile.LastWoWVersion ) ~= "number" ) or ( profile.LastWoWVersion < 30000 ) ) then
        --Upgrading from pre-wotlk, need to convert itemIDs in favorites and zone favorites to spellIDs
        local newList = TR:GetTable();
        local ItemInSet = PetAndMountDatabase.ItemInSet;

        for itemID, value in pairs( profile.Favorites ) do
            local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

            if( newID ) then
                itemID = tonumber( newID );
            end--if
            newList[itemID] = ( type( value ) == "number" ) and value or 2.0;
            if( newList[itemID] == 1.0 ) then
                newList[itemID] = nil;
            end--if
        end--for
        profile.Favorites = newList;
        newList = TR:GetTable();
        for zone, sets in pairs( profile.ZoneFavorites ) do
            newList[zone] = TR:GetTable();
            for set, mounts in pairs( sets ) do
                newList[zone][set] = TR:GetTable();
                for itemID, info in pairs( mounts ) do
                    local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

                    if( newID ) then
                        itemID = tonumber( newID );
                    end--if
                    newList[zone][set][itemID] = info;
                    if( not ( info.frequency or info.active or info.exclusive ) ) then
                        newList[zone][set][itemID] = nil;
                    end--if
                end--for
            end--for
        end--for
        profile.ZoneFavorites = newList;
    end--if
    if( ( type( profile.LastVersion ) == "table" ) and ( profile.LastVersion.major == 2 ) ) then
        --updating from 2.x, need to convert the zone favorites type names from set names to Major.Minor.Speed
        local newList = TR:GetTable();

        for zone, sets in pairs( profile.ZoneFavorites ) do
            newList[zone] = TR:GetTable();
            for set, mounts in pairs( sets ) do
                local subtype = set:find( "Passenger" ) and "Passenger" or "Normal";
                local speed = 0;
                local maintype = "Ground";

                if( set:find( "Ground" ) ) then
                    maintype = "Ground";
                    if( set:find( "Slow" ) ) then
                        speed = 0;
                    elseif( set:find( "Fast" ) ) then
                        speed = 60;
                    elseif( set:find( "VeryFast" ) ) then
                        speed = 100;
                    end--if
                elseif( set:find( "Flying" ) ) then
                    maintype = "Flying";
                    if( set:find( "Slow" ) ) then
                        speed = 0;
                    elseif( set:find( "Fast" ) ) then
                        speed = 60;
                    elseif( set:find( "VeryFast" ) ) then
                        speed = 200;
                    elseif( set:find( "ExtremelyFast" ) ) then
                        speed = 310;
                    end--if
                elseif( set:find( "Aquatic" ) ) then
                    maintype = "Aquatic";
                    if( set:find( "Slow" ) ) then
                        speed = 0;
                    elseif( set:find( "Fast" ) ) then
                        speed = 60;
                    elseif( set:find( "VeryFast" ) ) then
                        speed = 100;
                    end--if
                end--if
                newList[zone][("%s.%s.%i"):format( maintype, subtype, speed )] = mounts;
            end--if
        end--for
        profile.ZoneFavorites = newList;
    end--if
    profile.LastWoWVersion = LibLordFarlander.GetTOCVersion();
    profile.LastVersion = self.version;

    self.Dirty = true
    self:GetPlayersMounts();
    self:ChooseRandomMounts();
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:OnProfileChanged()

function Coconuts:OnEnable()
    --Register the Events
    local SpecialEvents_Mount = LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0" );
    local SpecialEvents_Skill = LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" );
    local LibUI = LibStub( "LibLordFarlander-UI-2.0" );

    self:RegisterEvent( "BAG_UPDATE",
        function( _, bagID )
            if( bagID and ( bagID >= 0 ) and ( bagID <= NUM_BAG_SLOTS ) ) then
                self.dirtyBags[bagID] = true;
                self.Dirty = true;
                if( not self.button:IsVisible() ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "COMPANION_UPDATE",
        function( _, type )
            if( not type ) then
                self.Dirty = true;
                self.dirtyBags[-255] = true;
                if( not self.button:IsVisible() ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "MODIFIER_STATE_CHANGED",
        function()
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "PLAYER_ALIVE",
        function()
            self.PlayerState.Mounted = ( SpecialEvents_Mount:PlayerMounted() ~= nil );
            self.PlayerState.Swimming = IsSwimming();
            self.PlayerState.InVehicle = UnitInVehicle( "player" );
            if( ( not self.Loaded ) or ( not self.PlayerState.RidingSkill ) ) then
                self.Loaded = true;
                self.PlayerState.RidingSkill = SpecialEvents_Skill:GetRidingSkillLevel();
                for bag = 0, NUM_BAG_SLOTS do
                    self.dirtyBags[bag] = true;
                end--for
                self.dirtyBags[-254] = true;
                self.dirtyBags[-255] = true;
                self:GetPlayersMounts();
                self:ChooseRandomMounts();
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "PLAYER_LEVEL_UP",
        function( _, level )
            -- todo?
        end );
    self:RegisterEvent( "PLAYER_REGEN_DISABLED",
        function()
            if( self.db.profile.Button.AutoHideButton and ( not ( self.db.profile.DefaultToFormInCombat and self.CurrentForm.Ground ) ) ) then
                self.button:HideEx();
            elseif( self.db.profile.DefaultToFormInCombat and self.CurrentForm.Ground ) then
                self.button:SetFromListEntry( self.CurrentForm.Ground );
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_ENABLED",
        function()
            if( ( not self.Loaded ) or ( not self.PlayerState.RidingSkill ) ) then
                self.Loaded = true;
                self.PlayerState.RidingSkill = SpecialEvents_Skill:GetRidingSkillLevel();
                for bag = 0, NUM_BAG_SLOTS do
                    self.dirtyBags[bag] = true;
                end--for
                self.dirtyBags[-254] = true;
                self.dirtyBags[-255] = true;
                self:GetPlayersMounts();
                self:ChooseRandomMounts();
                self:SetButton( Coconuts:GetStateBasedOnModifiers() );
            end--if
            if( self.db.profile.Button.AutoHideButton ) then
                self.button:ShowEx();
            end--if
            self.button:OutOfCombatReset();
            self.PlayerState.InVehicle = UnitInVehicle( "player" );
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "UNIT_ENTERED_VEHICLE",
        function( _, unit )
            if( unit == "player" ) then
                self.PlayerState.InVehicle = true;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "UNIT_EXITED_VEHICLE",
        function( _, unit )
            if( unit == "player" ) then
                self.PlayerState.InVehicle = false;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "ZONE_CHANGED",
        function()
            self.PlayerState.InVehicle = UnitInVehicle( "player" );
            if( self.db.profile.ZoneFavorites[GetRealZoneText()] ) then
                self:ChooseRandomMounts( true );
            end--if
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );

    SpecialEvents_Mount.RegisterCallback( self, "Mounted",
        function()
            if( not self.PlayerState.Mounted ) then
                self.PlayerState.Mounted = true;
                self:ChooseRandomMounts();
            end--if
        end );
    SpecialEvents_Mount.RegisterCallback( self, "Dismounted",
        function()
            if( self.PlayerState.Mounted ) then
                self.PlayerState.Mounted = false;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ).RegisterCallback( self, "LearnedSpell",
        function( _, spell )
            local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spell ) );

            --Check for learning Cold Weather Flying
            --Check for learning Flight Masters Licence
            if( ( spellID ~= nil ) and
                ( ( spellID == coldWeatherFlying ) or
                  ( spellID == flightMastersLicence ) or
                  PetAndMountDatabase:SpellInSet( spellID, "Spells.Travel" ) ) ) then
                self.Dirty = true;
                for bag = 0, NUM_BAG_SLOTS do
                    self.dirtyBags[bag] = true;
                end--for
                self.dirtyBags[-254] = true;
                self.dirtyBags[-255] = true;
                if( self.button and not( self.button:IsVisible() ) ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    SpecialEvents_Skill.RegisterCallback( self, "LearnedSkill_" .. ( ( LibLordFarlander.GetTOCVersion() >= 40000 ) and "Riding" or L["Riding"] ),
        function( _, level )
            self.PlayerState.RidingSkill = level;
            self:SetButton( self:GetStateBasedOnModifiers() );
            self:ResetMounts();
        end );
    SpecialEvents_Skill.RegisterCallback( self, "SkillLeveled_" .. ( ( LibLordFarlander.GetTOCVersion() >= 40000 ) and "Riding" or L["Riding"] ),
        function( _, level )
            self.PlayerState.RidingSkill = level;
            self:SetButton( self:GetStateBasedOnModifiers() );
            self:ResetMounts();
        end );

    if( not self.button ) then
        LibUI.SetupForUI( self );
        self.button = LibUI.CreateButton( self, self.name, [[Interface\AddOns\Coconuts\Img\Coconuts]], self.db.profile.ButtonGroup, self.SkinChanged, self );
        ItemList.AttachToButton( self.button );
        self.button:SetMovable( not self.db.profile.Button.LockButton );
        self.button:HookScript( "OnClick",
            function( self, mouse )
                if( ( mouse == "RightButton" ) and ( self.class.mode == "Enchant" ) ) then
                    self.class:FindNextMountEnchant();
                end--if
            end );
    end--if

    --Check if upgrading from Mount
    if( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ) ) then
        SetBindingClick( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ), self.button:GetName(), "LeftButton" );
        SetBindingClick( nil, "Mount_Button", "LeftButton" );
    end--if

    self:OnProfileChanged();
end--Coconuts:OnEnable();

function Coconuts:OnDisable()
    LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0" ).UnregisterAll( self );
    LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ).UnregisterAll( self );
    LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" ).UnregisterAll( self );
end--Coconuts:OnDisable()

function Coconuts:AutoHideChange()
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:AutoHideChange()

function Coconuts:OnUpdate( elapsed )
    local Aquatic = self.Mounts.List.Aquatic;
    local Underwater = self.Mounts.List.Underwater;
    local PlayerState = self.PlayerState;

    PlayerState.Swimming = IsSwimming();
    if( type( GetWintergraspWaitTime ) == "function" ) then
        PlayerState.WintergraspBattle = ( GetWintergraspWaitTime() == nil );
    end--if
    if( self.Dirty ) then
        --Update the list of mount items if needed
        self.Dirty = false;        
        if( self:HasMountListChanged() ) then
            self:ChooseRandomMounts();
        end--if
    end--if
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:OnUpdate()

function Coconuts:GetStateBasedOnModifiers()
    local profile = self.db.profile;
    local IsModifierDown = LibStub( "LibLordFarlander-UI-2.0" ).IsModifierDown;
    local IsFlyingModifierDown = IsModifierDown( profile.FlyingModifier );
    local IsPassengerModifierDown = IsModifierDown( profile.PassengerModifier );
    local IsMountItemModifierDown = IsModifierDown( profile.MountItemModifier );
    local IsNextBestModifierDown = IsModifierDown( profile.NextBestModifier );
    local Mounts = self.Mounts.List;

    if( self.PlayerState.InVehicle ) then
        self.Mode = "VehicleExit";
        return nil;
    elseif( self.PlayerState.Mounted ) then
        if( profile.ShowModifierItems and IsMountItemModifierDown and self.CurrentMountItem ) then
            self.Mode = "Enchant";
            return self.CurrentMountItem;
        else
            self.Mode = "Dismount";
            return nil;
        end--if
    elseif( Mounts.Underwater.Normal.Current and IsUsableSpell( Mounts.Underwater.Normal.Current.spellID ) ) then
        self.Mode = "Mount";
        return Mounts.Underwater.Normal.Current;
    elseif( self.PlayerState.Swimming and
            ( ( profile.SuggestForms and self.CurrentForm.Aquatic and IsUsableSpell( self.CurrentForm.Aquatic.spellID ) ) or
              ( Mounts.Aquatic.Normal.Current and IsUsableSpell( Mounts.Aquatic.Normal.Current.spellID ) ) ) ) then
        self.Mode = "Mount";
        return ( profile.SuggestForms and self.CurrentForm.Aquatic ) or
               Mounts.Aquatic.Normal.Current;
    else
        --Pick a mount or travel form
        if( ( profile.FavorFlying and ( not IsFlyingModifierDown ) ) or
            ( ( not profile.FavorFlying ) and IsFlyingModifierDown ) ) then
            if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or
                ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
                if( IsNextBestModifierDown ) then
                    if( Mounts.Flying.Passenger.NextCurrent and IsUsableSpell( Mounts.Flying.Passenger.NextCurrent.spellID ) ) then
                        self.Mode = "Mount";
                        return Mounts.Flying.Passenger.NextCurrent;
                    end--if
                else
                    if( Mounts.Flying.Passenger.Current and IsUsableSpell( Mounts.Flying.Passenger.Current.spellID ) ) then
                        self.Mode = "Mount";
                        return Mounts.Flying.Passenger.Current;
                    end--if
                end--if
            end--if
            if( IsNextBestModifierDown ) then
                if( profile.SuggestForms and ( not profile.FlightFormDefault ) and
                    self.CurrentForm.Flying and IsUsableSpell( self.CurrentForm.Flying.spellID ) ) then
                    self.Mode = "Mount";
                    return self.CurrentForm.Flying;
                end--if
                if( Mounts.Flying.Normal.NextCurrent and IsUsableSpell( Mounts.Flying.Normal.NextCurrent.spellID ) ) then
                    self.Mode = "Mount";
                    return Mounts.Flying.Normal.NextCurrent;
                end--if
            else
                if( profile.SuggestForms and ( not profile.FlightFormDefault ) and
                    self.CurrentForm.Flying and IsUsableSpell( self.CurrentForm.Flying.spellID ) ) then
                    self.Mode = "Mount";
                    return self.CurrentForm.Flying;
                end--if
                if( Mounts.Flying.Normal.Current and IsUsableSpell( Mounts.Flying.Normal.Current.spellID ) ) then
                    self.Mode = "Mount";
                    return Mounts.Flying.Normal.Current;
                end--if
            end--if
        end--if
        if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or
            ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
            if( IsNextBestModifierDown ) then
                if( ( profile.SuggestForms and self.CurrentForm.Ground and IsUsableSpell( self.CurrentForm.Ground.spellID ) ) or
                    ( Mounts.Ground.Passenger.NextCurrent and IsUsableSpell( Mounts.Ground.Passenger.NextCurrent.spellID ) ) ) then
                    self.Mode = "Mount";
                    return ( profile.SuggestForms and self.CurrentForm.Ground ) or Mounts.Ground.Passenger.NextCurrent.spellID;
                end--if
            else
                if( Mounts.Ground.Passenger.Current and IsUsableSpell( Mounts.Ground.Passenger.Current.spellID ) ) then
                    self.Mode = "Mount";
                    return Mounts.Ground.Passenger.Current;
                end--if
            end--if
        end--if
        if( IsNextBestModifierDown ) then
            if( ( profile.SuggestForms and self.CurrentForm.Ground and IsUsableSpell( self.CurrentForm.Ground.spellID ) ) or
                ( Mounts.Ground.Normal.NextCurrent and IsUsableSpell( Mounts.Ground.Normal.NextCurrent.spellID ) ) ) then
                self.Mode = "Mount";
                return ( profile.SuggestForms and self.CurrentForm.Ground ) or Mounts.Ground.Normal.NextCurrent;
            end--if
        else
            if( Mounts.Ground.Normal.Current and IsUsableSpell( Mounts.Ground.Normal.Current.spellID ) ) then
                self.Mode = "Mount";
                return Mounts.Ground.Normal.Current;
            end--if
        end--if
    end--if
    if( Mounts.Ground.Normal.Current or
        Mounts.Ground.Passenger.Current or
        Mounts.Flying.Normal.Current or
        Mounts.Flying.Passenger.Current or
        Mounts.Aquatic.Normal.Current or
        Mounts.Underwater.Normal.Current or
        self.CurrentForm.Ground or
        self.CurrentForm.Flying or
        self.CurrentForm.Aquatic ) then
        self.Mode = "NotMountable";
    else
        self.Mode = "";
    end--if
    return nil;
end--Coconuts:GetStateBasedOnModifiers()

----------------------------
--+Mount/Inventory Functions--
function Coconuts.GetSetSpeed( pt3set )
    local iterator, handle = PetAndMountDatabase:IterateSet( pt3set );
    local item, speed = iterator( handle );

    if( item and speed ) then
        local SpeedVar = 1;

        if( pt3set:find( "Flying" ) ) then
            SpeedVar = 3;
        elseif( pt3set:find( "Aquatic" ) ) then
            SpeedVar = 2;
        else
            SpeedVar = 1;
        end--if

        speed = select( SpeedVar, PetAndMountDatabase:GetMountAttributes( speed ) );
        return speed;
    end--if
    return nil;
end--Coconuts.GetSetSpeed( pt3set )

function Coconuts:GetZoneMount( set, ZoneList, noFallBack )
    local returnVal = nil;

    if( ZoneList and LibLordFarlander.HasTableMembers( ZoneList ) ) then
        local profile = self.db.profile;
        local favorites = TR:GetTable();
        local exclusives = TR:GetTable();

        for mount, settings in pairs( ZoneList ) do
            if( settings.active ) then
                favorites[mount] = settings.frequency or 1.0;
            end--if
            if( settings.exclusive ) then
                exclusives[mount] = settings.frequency or 1.0;
            end--if
        end--if
        if( LibLordFarlander.HasTableMembers( favorites ) or LibLordFarlander.HasTableMembers( exclusives ) ) then
            local default = 1.0;

            if( LibLordFarlander.HasTableMembers( exclusives ) ) then
                TR:ReturnTable( favorites );
                favorites = exclusives;
                exclusives = nil;
                default = 0.0;
            else
                local profileFavorites = self.db.profile.Favorites;

                for _, item in pairs( set ) do
                    local itemID = item:GetID();

                    if( not favorites[itemID] ) then
                        favorites[itemID] = profileFavorites[itemID] or default;
                    end--if
                end--if
            end--if
            returnVal = self:GetRandomFromListWithWeighing( set, favorites, default );
        end--if
        if( returnVal ) then
            local temp = returnVal;

            returnVal = TR:GetTable();
            LibLordFarlander.CopyTable( returnVal, temp );
        end--if
        TR:ReturnTable( favorites );
        if( exclusives ) then
            TR:ReturnTable( exclusives );
        end--if
    end--if
    if( not( returnVal or noFallBack ) ) then
        returnVal = self:GetRandomFromListWithWeighing( set, self.db.profile.Favorites );
        if( returnVal ) then
            local temp = returnVal;

            returnVal = TR:GetTable();
            LibLordFarlander.CopyTable( returnVal, temp );
        end--if
    end--if
    return returnVal;
end--Coconuts:GetZoneMount( set, ZoneList, noFallBack )

function Coconuts:GetRandomFromListWithWeighing( set, favorites, default )
    if( not default ) then
        default = 1.0;
    end--if
    if( set and ( LibLordFarlander.TableCount( set ) > 1 ) ) then
        local PlayerItemCount = LibLordFarlander.PlayerItemCount;
        local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
        local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
        local HasPlayerSpell = LibLordFarlander.HasPlayerSpell;
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local db = self.db;        
        local profile = db.profile;
        local FindInBags = ItemList.FindInBags;
        local opSuggestOnlySoulbound = profile.SuggestOnlySoulbound;
        local opFavorMorePassengers = profile.FavorMorePassengers;
        local total = 0.0;
        local useset = TR:GetTable();
        local randomValue;
        local amountOfItems = 0;
        local index = 0;
        local item = nil;
        local curTime = time();
        local FirstSeen = db.char.FirstSeen;
        local BoostNew = profile.BoostNew;
        local BoostNewTime = profile.BoostNewTime;
        local tempFavorites = TR:GetTable();

        for _, item in pairs( set ) do
            local reagented = nil;
            local itemID = item:GetID();
            if( itemID ) then
                local howOften = ( favorites[itemID] or default ) * ( opFavorMorePassengers and ( ( item.Passengers or 0.0 ) + 1.0 ) or 1.0 );
                local BagItem = ( itemID > 0 ) and FindInBags( itemID ) or nil;

                if( ( howOften > 0.0 ) and BoostNew and FirstSeen[itemID] and ( curTime - FirstSeen[itemID] <= BoostNewTime ) ) then
                    howOften = howOften + 20.0;
                end--if
                if( ( howOften > 0.0 ) and ( ( itemID < 0 ) or ( PlayerItemCount( itemID ) > 0 ) ) and ( ( itemID < 0 ) or ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) ) ) then
                    tinsert( useset, item );
                    total = total + howOften;
                    tempFavorites[itemID] = howOften;
                    amountOfItems = amountOfItems + 1;
                else
                    tempFavorites[itemID] = 0.0;
                end--if
            end--if
        end--for
        randomValue = ( math.random() + 0.05 ) * total;
        if( randomValue > total ) then
            randomValue = total;
        end--if
        repeat
            index = index + 1;
            item = useset[index];
            if( item ) then
                randomValue = randomValue - tempFavorites[item:GetID()];
            end--if
        until ( not item ) or ( randomValue <= 0 ) or ( index >= amountOfItems );
        table.wipe( useset );
        TR:ReturnTable( useset );
        TR:ReturnTable( tempFavorites );
        return item;
    else
        for _, item in pairs( set ) do
            return item;
        end--if
    end--if
end--Coconuts:GetRandomFromListWithWeighing( set, favorites, default )

function Coconuts:ChooseRandomMounts( noFallBack )
    local ZoneList = self.db.profile.ZoneFavorites[GetRealZoneText()];
    local Mounts = self.Mounts.List;

    Mounts.Ground.Normal.Current = self:GetZoneMount( self:GetPetsList( Mounts.Ground.Normal.Best, Mounts.Ground.Normal ), ZoneList and ZoneList[Mounts.Ground.Normal.Best], noFallBack ) or Mounts.Ground.Normal.Current;
    Mounts.Ground.Normal.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Ground.Normal.NextBest, Mounts.Ground.Normal ), ZoneList and ZoneList[Mounts.Ground.Normal.NextBest], noFallBack ) or Mounts.Ground.Normal.NextCurrent;
    Mounts.Flying.Normal.Current = self:GetZoneMount( self:GetPetsList( Mounts.Flying.Normal.Best, Mounts.Flying.Normal ), ZoneList and ZoneList[Mounts.Flying.Normal.Best], noFallBack ) or Mounts.Flying.Normal.Current;
    Mounts.Flying.Normal.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Flying.Normal.NextBest, Mounts.Flying.Normal ), ZoneList and ZoneList[Mounts.Flying.Normal.NextBest], noFallBack ) or Mounts.Flying.Normal.NextCurrent;
    Mounts.Aquatic.Normal.Current = self:GetZoneMount( self:GetPetsList( Mounts.Aquatic.Normal.Best, Mounts.Aquatic.Normal ), ZoneList and ZoneList[Mounts.Aquatic.Normal.Best], noFallBack ) or Mounts.Aquatic.Normal.Current;
    Mounts.Aquatic.Normal.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Aquatic.Normal.NextBest, Mounts.Aquatic.Normal ), ZoneList and ZoneList[Mounts.Aquatic.Normal.NextBest], noFallBack ) or Mounts.Aquatic.Normal.NextCurrent;
    Mounts.Underwater.Normal.Current = self:GetZoneMount( self:GetPetsList( Mounts.Underwater.Normal.Best, Mounts.Underwater.Normal ), ZoneList and ZoneList[Mounts.Underwater.Normal.Best], noFallBack ) or Mounts.Underwater.Normal.Current;
    Mounts.Underwater.Normal.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Underwater.Normal.NextBest, Mounts.Underwater.Normal ), ZoneList and ZoneList[Mounts.Underwater.Normal.NextBest], noFallBack ) or Mounts.Underwater.Normal.NextCurrent;

    Mounts.Ground.Passenger.Current = self:GetZoneMount( self:GetPetsList( Mounts.Ground.Passenger.Best, Mounts.Ground.Passenger ), ZoneList and ZoneList[Mounts.Ground.Passenger.Best], noFallBack ) or Mounts.Ground.Passenger.Current;
    Mounts.Ground.Passenger.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Ground.Passenger.NextBest, Mounts.Ground.Passenger ), ZoneList and ZoneList[Mounts.Ground.Passenger.NextBest], noFallBack ) or Mounts.Ground.Passenger.NextCurrent;
    Mounts.Flying.Passenger.Current = self:GetZoneMount( self:GetPetsList( Mounts.Flying.Passenger.Best, Mounts.Flying.Passenger ), ZoneList and ZoneList[Mounts.Flying.Passenger.Best], noFallBack ) or Mounts.Flying.Passenger.Current;
    Mounts.Flying.Passenger.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Flying.Passenger.NextBest, Mounts.Flying.Passenger ), ZoneList and ZoneList[Mounts.Flying.Passenger.NextBest], noFallBack ) or Mounts.Flying.Passenger.NextCurrent;
    Mounts.Aquatic.Passenger.Current = self:GetZoneMount( self:GetPetsList( Mounts.Aquatic.Passenger.Best, Mounts.Aquatic.Passenger ), ZoneList and ZoneList[Mounts.Aquatic.Passenger.Best], noFallBack ) or Mounts.Aquatic.Passenger.Current;
    Mounts.Aquatic.Passenger.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Aquatic.Passenger.NextBest, Mounts.Aquatic.Passenger ), ZoneList and ZoneList[Mounts.Aquatic.Passenger.NextBest], noFallBack ) or Mounts.Aquatic.Passenger.NextCurrent;
    Mounts.Underwater.Passenger.Current = self:GetZoneMount( self:GetPetsList( Mounts.Underwater.Passenger.Best, Mounts.Underwater.Passenger ), ZoneList and ZoneList[Mounts.Underwater.Passenger.Best], noFallBack ) or Mounts.Underwater.Passenger.Current;
    Mounts.Underwater.Passenger.NextCurrent = self:GetZoneMount( self:GetPetsList( Mounts.Underwater.Passenger.NextBest, Mounts.Underwater.Passenger ), ZoneList and ZoneList[Mounts.Underwater.Passenger.NextBest], noFallBack ) or Mounts.Underwater.Passenger.NextCurrent;

    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:ChooseRandomMounts()

function Coconuts.GetBestSets( sets )
    local Best = nil;
    local BestName = "";
    local NextBest = nil;
    local NextBestName = "";

    for name, set in pairs( sets ) do
        if( Best == nil ) then
            Best = set;
            BestName = name;
        else
            if( set.speed > Best.speed ) then
                if( ( NextBest == nil ) or ( Best.speed > NextBest.speed ) ) then
                    NextBest = Best;
                    NextBestName = BestName;
                end--if
                Best = set;
                BestName = name;
            elseif( ( NextBest == nil ) or ( set.speed > NextBest.speed ) ) then
                NextBest = set;
                NextBestName = name;
            end--if
        end--if
    end--for
    if( not NextBest ) then
        NextBestName = BestName;
    end--if
    return BestName, NextBestName;
end--Coconuts.GetBestSets()

function Coconuts.addToSet( dest, numspeed, bag, slot, itemID, passengers, skills )
    local item;
    
    if( not dest[numspeed] ) then
        dest[numspeed] = TR:GetTable( "bags", TR:GetTable(), "speed", numspeed );
    end--if
    if( not dest[numspeed].bags[bag] ) then
        dest[numspeed].bags[bag] = TR:GetTable();
    end--if
    if( itemID < 0 ) then
        item = ItemList.AddSpellToList( dest[numspeed].bags[bag], itemID * -1 );
    else
        item = ItemList.AddBagItemToList( dest[numspeed].bags[bag], bag, slot );
    end--if
    if( item ) then
        if( skills ) then
            for name, skill in pairs( skills ) do
                item[name] = skill;
            end--for
        end--if
        item.Passengers = passengers;
    end--if
end--Coconuts.addToSet( dest, numspeed, bag, slot, itemID, passengers, skills )

function Coconuts:AddIfMount( groundDest, groundPassengerDest, flyingDest, flyingPassengerDest, aquaticDest, aquaticPassengerDest, underwaterDest, underwaterPassengerDest, bag, slot, itemID, ridingSkill )
    local speed = PetAndMountDatabase:ItemInSet( itemID, "PetAndMountDatabase.Mounts" );

    if( speed ) then
        local numspeed;
        local groundspeed, aquaticspeed, flyingspeed, passengers, skills, underwaterspeed = PetAndMountDatabase:GetMountAttributes( speed );

--            self:dprint( ("%i;%s"):format( itemID, speed ) );
        if( groundspeed and ( groundspeed > 0 ) and ( ( flyingspeed == 0 ) or aquaticspeed ~= 0 ) ) then
            -- Don't scale walking speed mounts
            if( groundspeed > 100 ) then
                if( ridingSkill >= 150 ) then
                    groundspeed = 200;
                elseif( ridingSkill >= 75 ) then
                    groundspeed = 160;
                end--if
            end--if
            self.addToSet( groundDest, groundspeed, bag, slot, itemID, passengers, skills );
            if( passengers and ( passengers > 1 ) ) then
                self.addToSet( groundPassengerDest, groundspeed, bag, slot, itemID, passengers, skills );
            end--if
        end--if
        if( flyingspeed and ( flyingspeed > 0 ) ) then
            if( ridingSkill >= 375 ) then
                flyingspeed = 410;
            elseif( ridingSkill >= 300 ) then
                flyingspeed = 380;
            elseif( ridingSkill >= 225 ) then
                flyingspeed = 250;
            end--if
            self.addToSet( flyingDest, flyingspeed, bag, slot, itemID, passengers, skills );
            if( passengers and ( passengers > 1 ) ) then
                self.addToSet( flyingPassengerDest, flyingspeed, bag, slot, itemID, passengers, skills );
            end--if
        end--if
        if( underwaterspeed and ( underwaterspeed > 0 ) ) then
            self.addToSet( underwaterDest, underwaterspeed, bag, slot, itemID, passengers, skills );
            if( aquaticspeed and aquaticspeed > 0 ) then
                self.addToSet( aquaticDest, aquaticspeed, bag, slot, itemID, passengers, skills );
            end--if
            if( passengers and ( passengers > 1 ) ) then
                self.addToSet( underwaterPassengerDest, underwaterspeed, bag, slot, itemID, passengers, skills );
                if( aquaticspeed and aquaticspeed > 0 ) then
                    self.addToSet( aquaticPassengerDest, aquaticspeed, bag, slot, itemID, passengers, skills );
                end--if
            end--if
        elseif( aquaticspeed and ( aquaticspeed > 100 ) ) then
            self.addToSet( aquaticDest, aquaticspeed, bag, slot, itemID, passengers, skills );
            if( passengers and ( passengers > 1 ) ) then
                self.addToSet( aquaticPassengerDest, aquaticspeed, bag, slot, itemID, passengers, skills );
            end--if
        end--if
        return true;
    end--if
    return false;
end--Coconuts:AddIfMount( groundDest, groundPassengerDest, flyingDest, flyingPassengerDest, aquaticDest, aquaticPassengerDest, bag, slot, itemID, ridingSkill )

function Coconuts.GetTravelForm( set )
    local iterator, handle = PetAndMountDatabase:IterateSet( ("Spells.Travel.%s"):format( set ) );
    local spell, _ = iterator( handle );

    while spell do
        spell = spell * -1;
        if( LibLordFarlander.HasPlayerSpell( spell ) ) then
            return ItemList.BuildSpell( spell );
        end--if
        spell, _ = iterator( handle );
    end--for
    return nil;
end--Coconuts.GetTravelForm( set )

function Coconuts.CleanupBags( sets, bag )
    for _, set in pairs( sets ) do
        if( set.bags ) then
            if( set.bags[bag] ) then
                TR:ReturnTable( set.bags[bag] );
            end--if
            set.bags[bag] = TR:GetTable();
        end--if
    end--for
end--Coconuts.CleanupBags( set )

--Gets all your pets from your inventory.
function Coconuts:GetPlayersMounts()
    -- First do ground mountItems, best first
    local Mounts = self.Mounts.List;
    local db = self.db;
    local profile = db.profile;
    local opSuggestOnlySoulbound = profile.SuggestOnlySoulbound;
    local dirtyBags = self.dirtyBags;
    local Ground = Mounts.Ground;
    local Aquatic = Mounts.Aquatic;
    local Underwater = Mounts.Underwater;
    local Flying = Mounts.Flying;
    local GroundSets = Ground.Normal.Sets;
    local FlyingSets = Flying.Normal.Sets;
    local AquaticSets = Aquatic.Normal.Sets;
    local UnderwaterSets = Underwater.Normal.Sets;
    local GroundPassengerSets = Ground.Passenger.Sets;
    local FlyingPassengerSets = Flying.Passenger.Sets;
    local AquaticPassengerSets = Aquatic.Passenger.Sets;
    local UnderwaterPassengerSets = Underwater.Passenger.Sets;
    local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
    local PlayerMeetsMountRequirements = self.PlayerMeetsMountRequirements;
    local AddIfMount = self.AddIfMount;
    local AddIfVariableMount = self.AddIfVariableMount;
    local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
    local CleanupBags = self.CleanupBags;
    local FirstSeen = db.char.FirstSeen;
    local curTime = time();
    local added = false;
    local ridingSkill;

    self.CurrentMountItem = nil;
    Ground.Normal.Best = 0;
    Ground.Normal.NextBest = 0;
    Ground.Passenger.Best = 0;
    Ground.Passenger.NextBest = 0;
    Flying.Normal.Best = 0;
    Flying.Normal.NextBest = 0;
    Flying.Passenger.Best = 0;
    Flying.Passenger.NextBest = 0;
    Aquatic.Normal.Best = 0;
    Aquatic.Normal.NextBest = 0;
    Aquatic.Passenger.Best = 0;
    Aquatic.Passenger.NextBest = 0;
    Underwater.Normal.Best = 0;
    Underwater.Normal.NextBest = 0;
    Underwater.Passenger.Best = 0;
    Underwater.Passenger.NextBest = 0;

    if( not self.PlayerState.RidingSkill ) then
        self.PlayerState.RidingSkill = LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" ):GetRidingSkillLevel();
    end--if
    
    ridingSkill = self.PlayerState.RidingSkill or 0;

    for bag = 0, NUM_BAG_SLOTS do
        if( dirtyBags[bag] ) then
            CleanupBags( GroundSets, bag );
            CleanupBags( FlyingSets, bag );
            CleanupBags( AquaticSets, bag );
            CleanupBags( UnderwaterSets, bag );
            CleanupBags( GroundPassengerSets, bag );
            CleanupBags( FlyingPassengerSets, bag );
            CleanupBags( AquaticPassengerSets, bag );
            CleanupBags( UnderwaterPassengerSets, bag );
            for slot = 1, GetContainerNumSlots( bag ) do
                local itemLink = GetContainerItemLink( bag, slot );

                if( itemLink ) then
                    local itemID = GetItemIDFromLink( itemLink );

                    if( itemID and PlayerMeetsMountRequirements( self, itemID ) and ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( bag, slot ) ) ) then
                        added = AddIfMount( self, GroundSets, GroundPassengerSets, FlyingSets, FlyingPassengerSets, AquaticSets, AquaticPassengerSets, UnderwaterSets, UnderwaterPassengerSets, bag, slot, itemID, ridingSkill );
                        if( added ) then
                            FirstSeen[itemID] = curTime;
                        end--if
                    end--if
                end--if
            end--for
        end--if
    end--for

    if( dirtyBags[-255] ) then
        local mounts = PetAndMountDatabase:GetPlayersCompanions( "MOUNT" );

        CleanupBags( GroundSets, -255 );
        CleanupBags( FlyingSets, -255 );
        CleanupBags( AquaticSets, -255 );
        CleanupBags( UnderwaterSets, -255 );
        CleanupBags( GroundPassengerSets, -255 );
        CleanupBags( FlyingPassengerSets, -255 );
        CleanupBags( AquaticPassengerSets, -255 );
        CleanupBags( UnderwaterPassengerSets, -255 );

        for id, mount in pairs( mounts ) do
            local spellID = mount.spellID * -1;

            added = AddIfMount( self, GroundSets, GroundPassengerSets, FlyingSets, FlyingPassengerSets, AquaticSets, AquaticPassengerSets, UnderwaterSets, UnderwaterPassengerSets, -255, nil, spellID, ridingSkill );
            if( not added ) then
                --we KNOW this a mount, but the database is out of date and doesn't have it
                --we'll have to use the tooltip to figure out if it's flying or not and its speed
                --and if it can take passengers
                self:DetermineMount( mount, Mounts, FirstSeen, ridingSkill );
            elseif( not FirstSeen[spellID] ) then
                FirstSeen[spellID] = curTime;
            end--if
        end--for
        TR:ReturnTable( mounts );
    end--if
    
    if( self.reset ) then
        -- reset all mount data
        -- do it again
        self.reset = false;
        self:ResetMounts();
    else
        Ground.Normal.Best, Ground.Normal.NextBest = self.GetBestSets( GroundSets );
        Flying.Normal.Best, Flying.Normal.NextBest = self.GetBestSets( FlyingSets );
        Aquatic.Normal.Best, Aquatic.Normal.NextBest = self.GetBestSets( AquaticSets );
        Underwater.Normal.Best, Underwater.Normal.NextBest = self.GetBestSets( UnderwaterSets );
        Ground.Passenger.Best, Ground.Passenger.NextBest = self.GetBestSets( GroundPassengerSets );
        Flying.Passenger.Best, Flying.Passenger.NextBest = self.GetBestSets( FlyingPassengerSets );
        Aquatic.Passenger.Best, Aquatic.Passenger.NextBest = self.GetBestSets( AquaticPassengerSets );
        Underwater.Passenger.Best, Underwater.Passenger.NextBest = self.GetBestSets( UnderwaterPassengerSets );

        self.CurrentMountItemIndex = nil;
        self:FindNextMountEnchant();

        self.CurrentForm.Ground = self.GetTravelForm( "Ground" );
        self.CurrentForm.Flying = self.GetTravelForm( "Flying" );
        self.CurrentForm.Aquatic = self.GetTravelForm( "Aquatic" );
        self.dirtyBags = table.wipe( self.dirtyBags );
        if( not self.PlayerState.RidingSkill ) then
            self.dirtyBags[-255] = true;
        end--if
    end--if
end--Coconuts:GetPlayersMounts()

function Coconuts:DetermineMount( mount, Mounts, FirstSeen, ridingSkill )
    local spellID = mount.spellID * -1;
    local gratuity = LibStub( "LibGratuity-3.0", true );    

    self:Print( ("Could not identify %s [spellID=%i;creatureID=%i]!"):format( mount.name or "", mount.spellID, mount.modelID or 0 ) );
    gratuity:SetHyperlink( select( 1, GetSpellLink( mount.spellID ) ) );
    
    spellID = spellID * -1;

    local superType = "Normal";
    local mountType = "Ground";
    local speed = 160;
    local sets;
   
    if( gratuity:Find( L["[Cc]an only be [%w]+ed in [%w%s]-%."] ) ) then
        mountType = "Flying";
    end--if
    if( gratuity:Find( L["This is a fast [%w%s]-[%.!]"] ) ) then
        if( mountType == "Flying" ) then
            speed = 250;
        else
            speed = 160;        
        end--if
    end--if
    if( gratuity:Find( L["This is a very fast [%w%s]-[%.!]"] ) ) then
        if( mountType == "Flying" ) then
            speed = 380;
        else
            speed = 200;
        end--if
    end--if
    if( gratuity:Find( L["This is an extremely fast [%w%s]-[%.!]"] ) ) then
        speed = 410;
    end--if
    if( gratuity:Find( L["This [%w%s]- changes depending on your Riding skill%."] ) ) then
        --Variable speed, flying or ground only
        if( mountType == "Flying" ) then
            if( ridingSkill >= 375 ) then
                speed = 410;
            elseif( ridingSkill >= 300 ) then
                speed = 380;
            else
                speed = 250;
            end--if
        else
            if( ridingSkill >= 150 ) then
                speed = 200;
            else
                speed = 160;
            end--if
        end--if
    end--if
    if( gratuity:Find( L["can carry (%d+) passengers"] ) or gratuity:Find( L["complete with two buddies"] ) or gratuity:Find( L["Allows [%w%s]- passengers to board."] ) ) then
        superType = "Passenger";
    end--if
    if( gratuity:Find( L["This [%w%s]- changes depending on your Riding skill and location%."] ) ) then
        --whoo!  This mount is a ground mount in Azeroth, flying otherwise
        --Also the speed depends on the player's riding skill!
        if( ridingSkill >= 225 ) then
            if( ridingSkill >= 375 ) then
                speed = 410;
            elseif( ridingSkill >= 300 ) then
                speed = 380;
            elseif( ridingSkill >= 225 ) then
                speed = 250;
            end--if
            mountType = "Flying";
            --Also add to very fast ground mounts
            sets = Mounts.Ground.Normal.Sets;
            if( sets[200] == nil ) then
                sets[200] = { bags = TR:GetTable(), ["speed"] = 200 };
            end--if
            if( not sets[200].bags[-255] ) then
                sets[200].bags[-255] = TR:GetTable();
            end--if
            ItemList.AddSpellToList( sets[200].bags[-255], spellID );
        elseif( ridingSkill >= 150 ) then
            speed = 200;
            mountType = "Ground";
        elseif( ridingSkill >= 75 ) then
            speed = 160;
            mountType = "Ground";
        end--if
    end--if
    sets = Mounts[mountType][superType].Sets;
    if( sets[speed] == nil ) then
        sets[speed] = TR:GetTable( "bags", TR:GetTable(), "speed", speed );
    end--if
    if( not sets[speed].bags[-255] ) then
        sets[speed].bags[-255] = TR:GetTable();
    end--if
    if( not FirstSeen[spellID * -1] ) then
        FirstSeen[spellID * -1] = curTime;
    end--if
    ItemList.AddSpellToList( sets[speed].bags[-255], spellID );
    self:dprint( ("Added as %s %s [%i%%]"):format( superType, mountType, speed ) );
end--Coconuts:DetermineMount( mount, Mounts, FirstSeen, ridingSkill )

function Coconuts:PrintMountList( mountTable, typestring )
    if( #mountTable > 0 ) then
--        local zoneList = self:GetZoneList();
        local curTime = time();
        local profile = self.db.profile;
        local FirstSeen = profile.FirstSeen;
        local BoostNew = profile.BoostNew;
        local BoostNewTime = profile.BoostNewTime;  
    
        self:Print( L["%s mounts:"]:format( typestring ) );
        for _, mount in pairs( mountTable ) do
            local itemID = mount:GetID();
--            local howOften = zoneList[itemID] or 0.0;
            
--            if( ( howOften > 0.0 ) and BoostNew and FirstSeen[itemID] and ( curTime - FirstSeen[itemID] <= BoostNewTime ) ) then
--                howOften = howOften + 20.0;
--            end--if       
            self:Print( ("|T%s:0|t %s"):format( mount:GetTexture(), mount:GetLink() ) );
        end--for
    end--if
end--Coconuts:PrintMountList()

function Coconuts:ListMountType( mounttype, lists )
    local sorted = TR:GetTable();

    for name, _ in pairs( lists.Sets ) do
        tinsert( sorted, name );
    end--for
    table.sort( sorted );
    for _, name in pairs( sorted ) do
        self:PrintMountList( self:GetPetsList( name, lists ), L["%1$i%% speed %2$s"]:format( name - 100, mounttype ) );
    end--for
    TR:ReturnTable( sorted );
end--Coconuts.ListMountType( type, list )

function Coconuts:GetPetsList( set, parent )
    local returnVal = TR:GetTable();

    if( parent.Sets[set] and parent.Sets[set].bags ) then
        for _, bag in pairs( parent.Sets[set].bags ) do
            for _, mount in pairs( bag ) do
                tinsert( returnVal, mount );
            end--for
        end--for
    end--if
    return returnVal;
end--Coconuts:GetPetsList( set, parent )

function Coconuts.GetSetMountsList( sets, returnList, used )
    if( not used ) then
        used = TR:GetTable();
    end--if
    for _, set in pairs( sets ) do
        if( set.bags ) then
            for _, bag in pairs( set.bags ) do
                for _, mount in pairs( bag ) do
                    if( not used[mount:GetID()] ) then
                        tinsert( returnList, mount );
                        used[mount:GetID()] = true;
                    end--if
                end--for
            end--for
        end--if
    end--for
end--Coconuts.GetSetMountsList( set, returnList, used )

function Coconuts:GetAllMountsList()
    local returnList = TR:GetTable();
    local used = TR:GetTable();
    local Mounts = self.Mounts.List;

    self.GetSetMountsList( Mounts.Ground.Normal.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Flying.Normal.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Aquatic.Normal.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Underwater.Normal.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Ground.Passenger.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Flying.Passenger.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Aquatic.Passenger.Sets, returnList, used );
    self.GetSetMountsList( Mounts.Underwater.Passenger.Sets, returnList, used );
    TR:ReturnTable( used );
    return returnList;
end--Coconuts:GetAllMountsList()

function Coconuts:BuildZoneFavorites()
    TR:ReturnTable( self.Mounts.Zones );
    self.Mounts.Zones = TR:GetTable( "default", TR:GetTable() );

    local Zones = self.Mounts.Zones;
    local profile = self.db.profile;
    local profileFavorites = profile.Favorites;
    local ZoneFavorites = profile.ZoneFavorites;
    local default = Zones.default;

    local function ProcessZone( typename, name, subname, info, zone, list )
        local mountType = ("%s.%s.%i"):format( name, subname, typename );
        local mountFavoriteList = list[mountType];
        
        if( mountFavoriteList and LibLordFarlander.HasTableMembers( mountFavoriteList ) ) then
            local favorites = TR:GetTable();
            local exclusives = TR:GetTable();
            
            if( not Zones[zone] ) then
                Zones[zone] = TR:GetTable();
            end--if
            if( not Zones[zone][name] ) then
                Zones[zone][name] = TR:GetTable();
            end--if            
            for pet, settings in pairs( mountFavoriteList ) do
                if( settings.active ) then
                    favorites[pet] = settings.frequency or 1.0;
                end--if
                if( settings.exclusive ) then
                    exclusives[pet] = settings.frequency or 1.0;
                end--if
            end--if
            if( LibLordFarlander.HasTableMembers( exclusives ) ) then
                TR:ReturnTable( favorites );
                if( not Zones[zone][name][subname] ) then
                    Zones[zone][name][subname] = TR:GetTable();
                end--if
                Zones[zone][name][subname][typename] = exclusives;
            elseif( LibLordFarlander.HasTableMembers( favorites ) ) then
                TR:ReturnTable( exclusives );
                for _, item in pairs( mountList ) do
                    local id = item:GetID();

                    if( not favorites[id] ) then
                        favorites[id] = profileFavorites[id] or 1.0;
                    end--if
                end--if
                if( not Zones[zone][name][subname] ) then
                    Zones[zone][name][subname] = TR:GetTable();
                end--if
                Zones[zone][name][subname][typename] = favorites;
            else
                TR:ReturnTable( exclusives );
                TR:ReturnTable( favorites );
            end--if
        else
            list[mountType] = nil;
        end--if    
    end--function ProcessZone()
    
    for name, subtypes in pairs( self.Mounts.List ) do
        default[name] = TR:GetTable();

        local defaultMain = default[name];

        for subname, info in pairs( subtypes ) do
            local best = self:GetPetsList( info.Best, info );
            local nextbest = self:GetPetsList( info.NextBest, info );

            for zone, list in pairs( ZoneFavorites ) do
                for speed, _ in pairs( list ) do
                    ProcessZone( speed, name, subname, info, zone, list );                
                end--for
            end--for

            defaultMain[subname] = TR:GetTable();
            
            local defaultSub = defaultMain[subname];
            
            for speed, set in pairs( info.sets ) do
                defaultSub[speed][id] = profileFavorites[id] or 1.0;
            end--for
        end--for
    end--for
end--Coconuts:BuildZoneFavorites()

function Coconuts:ListMounts()
    local Mounts = self.Mounts.List;
    
    self:Print( L["List of your mounts:"] );

    self:ListMountType( L["ground"], Mounts.Ground.Normal );
    self:ListMountType( L["flying"], Mounts.Flying.Normal );
    self:ListMountType( L["swimming"], Mounts.Aquatic.Normal );
    self:ListMountType( L["underwater"], Mounts.Underwater.Normal );
    self:ListMountType( L["ground passenger"], Mounts.Ground.Passenger );
    self:ListMountType( L["flying passenger"], Mounts.Flying.Passenger );
    self:ListMountType( L["swimming passenger"], Mounts.Aquatic.Passenger );
    self:ListMountType( L["underwater passenger"], Mounts.Underwater.Passenger );
end--Coconuts:ListMounts()

function Coconuts:ResetMounts()
    for bag = 0, NUM_BAG_SLOTS do
        self.dirtyBags[bag] = true;
    end--for
    self.dirtyBags[-254] = true;
    self.dirtyBags[-255] = true;
    self.Mounts = {
        Zones = TR:GetTable(),
        List = {
            Ground = {
                Normal = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
                Passenger = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
            },
            Flying = {
                Normal = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
                Passenger = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
            },
            Aquatic = {
                Normal = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
                Passenger = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
            },
            Underwater = {
                Normal = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
                Passenger = {
                    Best = 0,
                    NextBest = 0,
                    Current = nil,
                    NextCurrent = nil,
                    Sets = { [0] = TR:GetTable( "speed", 0 ), },
                },
            },
        },
    };
    self:GetPlayersMounts();
    self:ChooseRandomMounts();
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:ResetMounts()
---Coconuts/Inventory Functions--
----------------------------

-------------------
--+Help Functions--
function Coconuts:SetOption_ResetMounts( info, val )
    local arg = info.arg;

    self.db.profile[info.arg.VarableName] = val;
    self:ResetMounts();
end--Coconuts:SetOption_ResetMounts( info, val )

function Coconuts:PlayerMeetsMountRequirements( itemID )
    local speed = PetAndMountDatabase:ItemInSet( itemID, "Mounts" );

    if( speed ) then
        local itemMinLevel = select( 5, GetItemInfo( itemID ) ) or 0;
        local ridingSkill = tonumber( select( 2, strsplit( "|", speed ) ) or 0 );

        return ( itemMinLevel <= UnitLevel( "player" ) ) and ( ridingSkill <= ( self.PlayerState.RidingSkill or 0 ) );
    end--if
    return false;
end--if

function Coconuts:HasMountListChanged()
    local Mounts = self.Mounts.List;
    local selfBestGround = Mounts.Ground.Normal.Sets[Mounts.Ground.Normal.Best];
    local selfBestFlying = Mounts.Flying.Normal.Sets[Mounts.Flying.Normal.Best];
    local selfBestAquatic = Mounts.Aquatic.Normal.Sets[Mounts.Aquatic.Normal.Best];
    local selfBestUnderwater = Mounts.Underwater.Normal.Sets[Mounts.Underwater.Normal.Best];
    local selfBestPassengerGround = Mounts.Ground.Passenger.Sets[Mounts.Ground.Passenger.Best];
    local selfBestPassengerFlying = Mounts.Flying.Passenger.Sets[Mounts.Flying.Passenger.Best];
    local selfBestPassengerAquatic = Mounts.Aquatic.Passenger.Sets[Mounts.Aquatic.Passenger.Best];
    local selfBestPassengerUnderwater = Mounts.Underwater.Passenger.Sets[Mounts.Underwater.Passenger.Best];
    local selfNextBestGround = Mounts.Ground.Normal.Sets[Mounts.Ground.Normal.NextBest];
    local selfNextBestFlying = Mounts.Flying.Normal.Sets[Mounts.Flying.Normal.NextBest];
    local selfNextBestAquatic = Mounts.Aquatic.Normal.Sets[Mounts.Aquatic.Normal.NextBest];
    local selfNextBestUnderwater = Mounts.Underwater.Normal.Sets[Mounts.Underwater.Normal.NextBest];
    local selfNextBestPassengerGround = Mounts.Ground.Passenger.Sets[Mounts.Ground.Passenger.NextBest];
    local selfNextBestPassengerFlying = Mounts.Flying.Passenger.Sets[Mounts.Flying.Passenger.NextBest];
    local selfNextBestPassengerAquatic = Mounts.Aquatic.Passenger.Sets[Mounts.Aquatic.Passenger.NextBest];
    local selfNextBestPassengerUnderwater = Mounts.Underwater.Passenger.Sets[Mounts.Underwater.Passenger.NextBest];
    local preGroundCount = #selfBestGround;
    local preFlyingCount = #selfBestFlying;
    local preAquaticCount = #selfBestAquatic;
    local preUnderwaterCount = #selfBestUnderwater;
    local prePassengerGroundCount = #selfBestPassengerGround;
    local prePassengerFlyingCount = #selfBestPassengerFlying;
    local prePassengerAquaticCount = #selfBestPassengerAquatic;
    local prePassengerUnderwaterCount = #selfBestPassengerUnderwater;
    local preNextGroundCount = #selfNextBestGround;
    local preNextFlyingCount = #selfNextBestFlying;
    local preNextAquaticCount = #selfNextBestAquatic;
    local preNextUnderwaterCount = #selfNextBestUnderwater;
    local preNextPassengerGroundCount = #selfNextBestPassengerGround;
    local preNextPassengerFlyingCount = #selfNextBestPassengerFlying;
    local preNextPassengerAquaticCount = #selfNextBestPassengerAquatic;
    local preNextPassengerUnderwaterCount = #selfNextBestPassengerUnderwater;
    local preBestGround = TR:GetTable();
    local preBestFlying = TR:GetTable();
    local preBestAquatic = TR:GetTable();
    local preBestUnderwater = TR:GetTable();
    local preBestPassengerGround =  TR:GetTable();
    local preBestPassengerFlying = TR:GetTable();
    local preBestPassengerAquatic = TR:GetTable();
    local preBestPassengerUnderwater = TR:GetTable();
    local preNextBestGround =  TR:GetTable();
    local preNextBestFlying = TR:GetTable();
    local preNextBestAquatic = TR:GetTable();
    local preNextBestUnderwater = TR:GetTable();
    local preNextBestPassengerGround =  TR:GetTable();
    local preNextBestPassengerFlying = TR:GetTable();
    local preNextBestPassengerAquatic = TR:GetTable();
    local preNextBestPassengerUnderwater = TR:GetTable();
    local returnVal = false;

    LibLordFarlander.CopyTable( preBestGround, selfBestGround );
    LibLordFarlander.CopyTable( preBestFlying, selfBestFlying );
    LibLordFarlander.CopyTable( preBestAquatic, selfBestAquatic );
    LibLordFarlander.CopyTable( preBestUnderwater, selfBestUnderwater );
    LibLordFarlander.CopyTable( preBestPassengerGround, selfBestPassengerGround );
    LibLordFarlander.CopyTable( preBestPassengerFlying, selfBestPassengerFlying );
    LibLordFarlander.CopyTable( preBestPassengerAquatic, selfBestPassengerAquatic );
    LibLordFarlander.CopyTable( preBestPassengerUnderwater, selfBestPassengerUnderwater );
    LibLordFarlander.CopyTable( preNextBestGround, selfNextBestGround );
    LibLordFarlander.CopyTable( preNextBestFlying, selfNextBestFlying );
    LibLordFarlander.CopyTable( preNextBestAquatic, selfNextBestAquatic );
    LibLordFarlander.CopyTable( preNextBestUnderwater, selfNextBestUnderwater );
    LibLordFarlander.CopyTable( preNextBestPassengerGround, selfNextBestPassengerGround );
    LibLordFarlander.CopyTable( preNextBestPassengerFlying, selfNextBestPassengerFlying );
    LibLordFarlander.CopyTable( preNextBestPassengerAquatic, selfNextBestPassengerAquatic );
    LibLordFarlander.CopyTable( preNextBestPassengerUnderwater, selfNextBestPassengerUnderwater );

    self:GetPlayersMounts();

    MountsBestGround = Mounts.Ground.Normal.Sets[Mounts.Ground.Normal.Best];
    MountsBestFlying = Mounts.Flying.Normal.Sets[Mounts.Flying.Normal.Best];
    MountsBestAquatic = Mounts.Aquatic.Normal.Sets[Mounts.Aquatic.Normal.Best];
    MountsBestUnderwater = Mounts.Aquatic.Normal.Sets[Mounts.Underwater.Normal.Best];
    MountsBestPassengerGround = Mounts.Ground.Passenger.Sets[Mounts.Ground.Passenger.Best];
    MountsBestPassengerFlying = Mounts.Flying.Passenger.Sets[Mounts.Flying.Passenger.Best];
    MountsBestPassengerAquatic = Mounts.Aquatic.Passenger.Sets[Mounts.Aquatic.Passenger.Best];
    MountsBestPassengerUnderwater = Mounts.Underwater.Passenger.Sets[Mounts.Underwater.Passenger.Best];
    MountsNextBestGround = Mounts.Ground.Normal.Sets[Mounts.Ground.Normal.NextBest];
    MountsNextBestFlying = Mounts.Flying.Normal.Sets[Mounts.Flying.Normal.NextBest];
    MountsNextBestAquatic = Mounts.Aquatic.Normal.Sets[Mounts.Aquatic.Normal.NextBest];
    MountsNextBestUnderwater = Mounts.Underwater.Normal.Sets[Mounts.Underwater.Normal.NextBest];
    MountsNextBestPassengerGround = Mounts.Ground.Passenger.Sets[Mounts.Ground.Passenger.NextBest];
    MountsNextBestPassengerFlying = Mounts.Flying.Passenger.Sets[Mounts.Flying.Passenger.NextBest];
    MountsNextBestPassengerAquatic = Mounts.Aquatic.Passenger.Sets[Mounts.Aquatic.Passenger.NextBest];
    MountsNextBestPassengerUnderwater = Mounts.Underwater.Passenger.Sets[Mounts.Underwater.Passenger.NextBest];

    if( not self.ReadyToInitialize ) then
        self.ReadyToInitialize = true;
        return true;
    end
    self.ReadyToInitialize = true;
    returnVal = not( LibLordFarlander.CompareTables( preBestGround, selfBestGround ) and
                     LibLordFarlander.CompareTables( preBestFlying, selfBestFlying ) and
                     LibLordFarlander.CompareTables( preBestAquatic, selfBestAquatic ) and
                     LibLordFarlander.CompareTables( preBestUnderwater, selfBestUnderwater ) and
                     -- next best
                     LibLordFarlander.CompareTables( preNextBestGround, selfNextBestGround ) and
                     LibLordFarlander.CompareTables( preNextBestFlying, selfNextBestFlying ) and
                     LibLordFarlander.CompareTables( preNextBestAquatic, selfNextBestAquatic ) and
                     LibLordFarlander.CompareTables( preNextBestUnderwater, selfNextBestUnderwater ) and
                     -- best passenger
                     LibLordFarlander.CompareTables( preBestPassengerGround, selfBestPassengerGround ) and
                     LibLordFarlander.CompareTables( preBestPassengerFlying, selfBestPassengerFlying ) and
                     LibLordFarlander.CompareTables( preBestPassengerAquatic, selfBestPassengerAquatic ) and
                     LibLordFarlander.CompareTables( preBestPassengerUnderwater, selfBestPassengerUnderwater ) and
                     -- next best passenger
                     LibLordFarlander.CompareTables( preNextBestPassengerGround, selfNextBestPassengerGround ) and
                     LibLordFarlander.CompareTables( preNextBestPassengerFlying, selfNextBestPassengerFlying ) and
                     LibLordFarlander.CompareTables( preNextBestPassengerAquatic, selfNextBestPassengerAquatic ) and
                     LibLordFarlander.CompareTables( preNextBestPassengerUnderwater, selfNextBestPassengerUnderwater ) );
    TR:ReturnTable( preBestGround );
    TR:ReturnTable( preBestFlying );
    TR:ReturnTable( preBestAquatic );
    TR:ReturnTable( preBestUnderwater );
    TR:ReturnTable( preBestPassengerGround );
    TR:ReturnTable( preBestPassengerFlying );
    TR:ReturnTable( preBestPassengerAquatic );
    TR:ReturnTable( preBestPassengerUnderwater );
    TR:ReturnTable( preNextBestGround );
    TR:ReturnTable( preNextBestFlying );
    TR:ReturnTable( preNextBestAquatic );
    TR:ReturnTable( preNextBestUnderwater );
    TR:ReturnTable( preNextBestPassengerGround );
    TR:ReturnTable( preNextBestPassengerFlying );
    TR:ReturnTable( preNextBestPassengerAquatic );
    TR:ReturnTable( preNextBestPassengerUnderwater );
    return returnVal;
end--Coconuts:HasMountListChanged()

function Coconuts:FindNextMountEnchant()
    local enchantOrder = self.db.profile.enchantOrder;
    local index = ( self.CurrentMountItemIndex or 0 ) + 1;
    local total = #enchantOrder;
    local count = 0;

    if( index > #enchantOrder ) then
        index = 1;
    end--if
    while( count < total ) do
        if( GetItemCount( enchantOrder[index] ) > 0 ) then
            self.CurrentMountItem = ItemList.BuildItem( enchantOrder[index] );
            self.CurrentMountItemIndex = index;
            return;
        end--if
        index = index + 1;
        if( index > #enchantOrder ) then
            index = 1;
        end--if
        count = count + 1;
    end--while
    self.CurrentMountItem = nil;
    self.CurrentMountItemIndex = nil;
end--Coconuts:FindNextMountEnchant()

function Coconuts:QuickSummon( mountID )
    local companionID = PetAndMountDatabase:FindCompanion( "MOUNT", mountID )
    if( companionID ) then
       CallCompanion( "MOUNT", companionID );
       self:ChooseRandomMounts();
       self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
end--Coconuts:QuickSummon( mountID )
---Help Functions--
-------------------

-----------------
--+UI Functions--
function Coconuts:SetTooltip( tooltip )
    if( self.Mode == "Dismount" ) then
        local chardb = self.db.profile;

        if( chardb.ModiferDismountFlying ) then
            if( chardb.ShowModifierItems ) then
                local notUsed = TR:GetTable();
                local used = self.db.profile.MountItemModifier;
                
                if( used ~= "Alt" ) then
                    tinsert( notUsed, L["Alt"] );
                end--if
                if( used ~= "Control" ) then
                    tinsert( notUsed, L["Control"] );
                end--if
                if( used ~= "Shift" ) then
                    tinsert( notUsed, L["Shift"] );
                end--if
                tooltip:SetText( L["Dismount (if flying hold down %s or %s and click to dismount)"]:format( notUsed[1], notUsed[2] ) );
                TR:ReturnTable( notUsed );
            else
                tooltip:SetText( L["Dismount (if flying hold down Shift, Control, or Alt and click to dismount)"] );
            end--if
        else
            tooltip:SetText( L["Dismount"] );
        end--if
    elseif( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) then
        ItemList.SetTooltipFromListEntry( tooltip, self.CurrentMount );
    elseif( self.Mode == "VehicleExit" ) then
        tooltip:SetText( L["Exit Vehicle"] );
    else
        tooltip:SetText( L["You have no mounts."] );
    end--if
end--Coconuts:SetTooltip( tooltip )

function Coconuts:SetGameTooltips()
    local mntButton = self.button;
    local mntTooltip = mntButton.tooltip;

    if( ( not self.keyBoundMode ) and ( not InCombatLockdown() ) ) then
        mntTooltip:SetOwner( mntButton, "RIGHT" );
        self:SetTooltip( mntTooltip );
    elseif( InCombatLockdown() ) then
        mntTooltip:SetText( L["Cannot mount, you are in combat!"] );
    end--if
    LibStub( "LibLordFarlander-UI-2.0" ).AfterSetTooltip( mntButton );
end--Coconuts:SetGameTooltips()

function Coconuts:SetButton( mountItem )
    if( not InCombatLockdown() ) then
        local mntButton = self.button;
        local chardb = self.db.profile;
        local ldbobj = self.ldbobj;

        if( chardb.Button.AutoHideButton and mountItem ) then
            mntButton:ShowEx();
        end--if
        if( self.Mode == "Dismount" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", chardb.ModiferDismountFlying and "/dismount [noflying][mod]" or "/dismount" );
            mntButton:SetImage( [[Interface\AddOns\Coconuts\Img\Dismount]] );

            ldbobj.icon = [[Interface\AddOns\Coconuts\Img\Dismount]];
            ldbobj.text = L["Dismount"];
        elseif( self.Mode == "VehicleExit" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", "/script VehicleExit();" );
            mntButton:SetImage( [[Interface\AddOns\Coconuts\Img\VehicleExit]] );

            ldbobj.icon = [[Interface\AddOns\Coconuts\Img\VehicleExit]];
            ldbobj.text = L["Exit Vehicle"];
        elseif( mountItem and ( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) ) then
            mntButton:Enable();
            local macro;
            
            self.CurrentMountIndex = PetAndMountDatabase:FindCompanion( "MOUNT", mountItem.spellID );
            if( mountItem.spellID and self.CurrentMountIndex ) then
                macro = ([[/script CallCompanion( "MOUNT", %i );]]):format( self.CurrentMountIndex );
            end--if
            mntButton:SetFromListEntry( mountItem, not LibLordFarlander.CompareTables( self.CurrentMount, mountItem ), nil, nil, macro );
        elseif( self.Mode == "NotMountable" ) then
            if( chardb.Button.AutoHideButton ) then
                mntButton:HideEx();
            end--if
            mntButton.cooldown:Hide();
            mntButton:SetAttribute( "*type1", "item" );
            mntButton:SetAttribute( "*bag1", nil );
            mntButton:SetAttribute( "*slot1", nil );
            mntButton:SetAttribute( "*item1", nil );
            mntButton:SetImage( [[Interface\AddOns\Coconuts\Img\Dismount]] );

            ldbobj.icon = [[Interface\AddOns\Coconuts\Img\Dismount]];
            ldbobj.text = L["Cannot mount here."];
        else
            if( chardb.Button.AutoHideButton ) then
                mntButton:HideEx();
            end--if
            mntButton.cooldown:Hide();
            mntButton:SetAttribute( "*type1", "item" );
            mntButton:SetAttribute( "*bag1", nil );
            mntButton:SetAttribute( "*slot1", nil );
            mntButton:SetAttribute( "*item1", nil );
            mntButton:SetImage( [[Interface\AddOns\Coconuts\Img\Coconuts]] );

            ldbobj.icon = [[Interface\AddOns\Coconuts\Img\Coconuts]];
            ldbobj.text = L["You have no mounts."];
        end--if
        self.CurrentMount = mountItem;
    end--if
end--Coconuts:SetButton( mountItem )
---UI Functions--
-----------------
